/*
 * このプログラムはvispackおよびvisionaire.bmsを参考に作成しました
 * 基本的な処理はvispackとほぼ同じです
 *
 * vispack
 * https://googledrive.com/host/0ByAtt-u9ddI8eDlNS0xuOUM5cTA/deponia_jp.zip
 *
 * Visionaire vis
 * http://aluigi.altervista.org/papers/bms/visionaire.bms
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include <zlib.h>

#include "file_utils.h"
#include "utils.h"
#ifdef WIN32
#include "my_printf.h"
#define	printf	my_printf
#endif


#ifdef WIN32
static const char PATH_SEP = '\\';
#else
static const char PATH_SEP = '/';
#endif


#ifndef REPLACE_VIS
enum run_mode_e {
	MODE_LIST = 0,
	MODE_TEST,
	MODE_EXTRACT
};
#endif	/* not REPLACE_VIS */

enum byte_order_e {
	BYTE_ORDER_UNKNOWN = 0,
	BYTE_ORDER_LITTLE_ENDIAN,
	BYTE_ORDER_BIG_ENDIAN
};

static const unsigned int MAX_VIS_INDEX = 99999;
static const unsigned int MAX_VIS_FILES = 99999 + 1;

static const int VIS_KEY_SIZE = 16;

static char *default_key_file = "vis_key.txt";

static char *empty[] = { NULL };
static char **vis_key_table = empty;

struct vis_file_info_t {
	long offset;
	long z_size;
	long size;
	unsigned int flag;
	const char *ext;
	unsigned char *data;
#ifdef REPLACE_VIS
	unsigned char *new_data;
	long new_z_size;
	long new_size;
#endif	/* REPLACE_VIS */
};

/*
 * flag: bit 0 - XML
 *       bit 1 - VISキーでマスク
 *       bit 3 - マスク画像
 *       bit 4 - XML(分割圧縮)
 */
#define IS_XML1(_f)				(((_f) & 0x01) != 0)
#define IS_VIS_KEY_MASKED(_f)	(((_f) & 0x02) != 0)
#define IS_MASKED_IMAGE(_f)		(((_f) & 0x08) != 0)
#define IS_XML2(_f)				(((_f) & 0x10) != 0)

static int verbose_level = 0;


static void usage(void);
#ifdef REPLACE_VIS
static int replace_vis(const char *in_file, const char *out_file, char **replace_dirs, int num_replace_dirs, int *vis_index, int num_vis_index, const char *ext, enum byte_order_e byte_order, const char *vis_key, int opt_much_index_only);
#if 0
static int get_block_size_and_end_mark(const struct vis_file_info_t *info, unsigned int *block_size, unsigned int *end_mark, enum byte_order_e byte_order);
#endif
#else	/* REPLACE_VIS */
static int extract_vis(const char *in_file, const char *out_dir, int *vis_index, int num_vis_index, const char *ext, enum byte_order_e byte_order, const char *vis_key, enum run_mode_e mode);
#endif	/* REPLACE_VIS */
static int check_vis_key(char *des, const char *src);
static int read_key_file(const char *filename);
static int read_vis_file(const char *filename, unsigned char **data, long *size);
#ifdef REPLACE_VIS
static int read_replace_files(struct vis_file_info_t *info, unsigned int num_files, const char *replace_dir, int *vis_index, int num_vis_index, const char *ext, int opt_much_index_only);
static int read_file(const char *filename, unsigned char **data, long *size);
static int write_vis_file(const char *filename, const unsigned char *vis_data, const struct vis_file_info_t *info, unsigned int num_files);
#else	/* REPLACE_VIS */
static int write_file(const char *filename, const unsigned char *data, long size);
#endif	/* REPLACE_VIS */
static unsigned int get_int4(const unsigned char *data, enum byte_order_e byte_order);
#ifdef REPLACE_VIS
static void put_int4(unsigned char *data, unsigned int value, enum byte_order_e byte_order);
#endif	/* REPLACE_VIS */
static void xor(unsigned char *data, long size, const char *key);
#if 0
static void lame_vis_fix(struct vis_file_info_t *info, int index, int *lame_vis);
#endif
static void get_ext(struct vis_file_info_t *info);
static int compare_index(const void *a, const void *b);


int main(int argc, char **argv)
{
	int num_cmd_params = 2;
	const char *program_file = *argv;
	const char *in_file = NULL;
#ifdef REPLACE_VIS
	const char *out_file = NULL;
#else	/* REPLACE_VIS */
	const char *out_dir = NULL;
#endif	/* REPLACE_VIS */
	char *key_file = NULL;
#ifdef REPLACE_VIS
	char **replace_dirs = NULL;
	int num_replace_dirs = 0;
	int opt_much_index_only = 0;
#else	/* REPLACE_VIS */
	enum run_mode_e mode = MODE_EXTRACT;
#endif	/* REPLACE_VIS */

	enum byte_order_e byte_order = BYTE_ORDER_UNKNOWN;
	const char *vis_key = NULL;
	char vis_key_buf[VIS_KEY_SIZE + 1];

	int num_vis_index = 0;
	int *vis_index = NULL;

	const char *ext = NULL;

	memset(vis_key_buf, 0, sizeof(vis_key_buf));

	while (*++argv != NULL) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "--verbose") == 0) {
				verbose_level++;
			} else if (strcmp(*argv, "-B") == 0 || strcmp(*argv, "--big-endian") == 0) {
				byte_order = BYTE_ORDER_BIG_ENDIAN;
			} else if (strcmp(*argv, "-L") == 0 || strcmp(*argv, "--little-endian") == 0) {
				byte_order = BYTE_ORDER_LITTLE_ENDIAN;
			} else if (strcmp(*argv, "-k") == 0 || strcmp(*argv, "--key") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: VISキーがありません\n");
					return 1;
				}
				if (check_vis_key(vis_key_buf, *argv) < 0) {
					printf("ERROR: 不正なVISキーです: %s\n", *argv);
					return 1;
				}
				vis_key = vis_key_buf;
			} else if (strcmp(*argv, "-K") == 0 || strcmp(*argv, "--key-file") == 0) {
				if (*++argv == NULL) {
					printf("ERRPR: キーファイルがありません\n");
					return 1;
				}
				key_file = *argv;
#ifdef REPLACE_VIS
			} else if (strcmp(*argv, "--much-index-only") == 0) {
				opt_much_index_only = 1;
#else	/* REPLACE_VIS */
			} else if (strcmp(*argv, "-l") == 0 || strcmp(*argv, "--list") == 0) {
				mode = MODE_LIST;
			} else if (strcmp(*argv, "-t") == 0 || strcmp(*argv, "--test") == 0) {
				mode = MODE_TEST;
#endif	/* REPLACE_VIS */
			} else if (strcmp(*argv, "-i") == 0 || strcmp(*argv, "--index") == 0) {
				char *endptr = NULL;
				long index = 0;

				if (*++argv == NULL) {
					printf("ERROR: インデックスがありません\n");
					return 1;
				}
				index = strtol(*argv, &endptr, 10);
				if (endptr == NULL || *endptr != '\0' || index < 0 || MAX_VIS_INDEX < index) {
					printf("ERROR: 不正なインデックスです: %s\n", *argv);
					return 1;
				}

				/* 登録済みであれば無視する */
				if (num_vis_index != 0) {
					int found = 0;
					int i;
					for (i = 0; i < num_vis_index; i++) {
						if (vis_index[i] == index) {
							found = 1;
							break;
						}
					}
					if (found) {
						continue;
					}
				}

				if (num_vis_index == 0) {
					vis_index = (int *)malloc(sizeof(int));
				} else {
					vis_index = (int *)realloc(vis_index, sizeof(int) * (num_vis_index + 1));
				}
				if (vis_index == NULL) {
					printf("ERROR: memory exhausted!\n");
					return 1;
				}
				vis_index[num_vis_index++] = (int)index;
			} else if (strcmp(*argv, "-e") == 0 || strcmp(*argv, "--ext") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: 拡張子がありません\n");
					return 1;
				}
				ext = *argv;
			} else {
				printf("ERROR: 不正なオプションです: %s\n", *argv);
				return 1;
			}
		} else {
#ifdef REPLACE_VIS
			if (num_cmd_params == 2) {
				in_file = *argv;
			} else if (num_cmd_params == 1) {
				out_file = *argv;
			} else {
				if (num_replace_dirs == 0) {
					replace_dirs = (char **)malloc(sizeof(char *));
				} else {
					replace_dirs = (char **)realloc(replace_dirs, sizeof(char *) * (num_replace_dirs + 1));
				}
				if (replace_dirs == NULL) {
					printf("ERROR: memory exhausted!\n");
					return 1;
				}
				replace_dirs[num_replace_dirs++] = *argv;
			}
#else	/* REPLACE_VIS */
			if (num_cmd_params <= 0) {
				printf("ERROR: コマンドラインパラメータが多すぎます: %s\n", *argv);
				return 1;
			}
			if (num_cmd_params == 2) {
				in_file = *argv;
			} else {
				out_dir = *argv;
			}
#endif	/* REPLACE_VIS */
			num_cmd_params--;
		}
	}

#ifdef REPLACE_VIS
	if (in_file == NULL || out_file == NULL || num_replace_dirs == 0)
#else	/* REPLACE_VIS */
	if (in_file == NULL)
#endif	/* REPLACE_VIS */
	{
		usage();
		return 1;
	}

	if (1 < num_vis_index) {
		qsort(vis_index, num_vis_index, sizeof(vis_index[0]), compare_index);
	}

	if (vis_key == NULL) {
		if (key_file == NULL) {
			/* デフォルトのVISキーファイルを読み込む */
			char *ptr = strrchr(program_file, '/');
#ifdef WIN32
			char *ptr2 = strrchr(program_file, '\\');
			if (ptr2 != NULL && (ptr == NULL || ptr < ptr2)) {
				ptr = ptr2;
			}
#endif
			if (ptr == NULL) {
				key_file = default_key_file;
			} else {
				size_t dirname_len = (ptr - program_file);
				size_t size =  dirname_len + 1 + strlen(default_key_file) + 1;
				if ((key_file = (char *)malloc(size)) == NULL) {
					printf("ERROR: memory exhausted!\n");
					return 1;
				}
				strncpy(key_file, program_file, dirname_len);
				key_file[dirname_len] = PATH_SEP;
				strcpy(key_file + dirname_len + 1, default_key_file);
			}
		}
		if (read_key_file(key_file) < 0) {
			printf("ERROR: VISキーファイルを読めませんでした: %s\n", key_file);
			return 1;
		}
	} else {
		if (key_file != NULL) {
			printf("WARNING: VISキーが指定されているときはキーファイルを読みません\n");
		}
	}

#ifdef REPLACE_VIS
	if (replace_vis(in_file, out_file, replace_dirs, num_replace_dirs, vis_index, num_vis_index, ext, byte_order, vis_key, opt_much_index_only) < 0) {
		printf("ERROR: エラーにより処理を中断しました\n");
		return 1;
	}

	printf("ファイルを置換しました\n");
#else	/* REPLACE_VIS */
	if (mode != MODE_EXTRACT && out_dir != NULL) {
		printf("WARNING: %sオプションを指定したときは出力ディレクトリは不要です\n", (mode == MODE_LIST ? "-l|--list" : "-t|--test"));
	}

	if (extract_vis(in_file, out_dir, vis_index, num_vis_index, ext, byte_order, vis_key, mode) < 0) {
		printf("ERROR: エラーにより処理を中断しました\n");
		return 1;
	}

	if (mode == MODE_EXTRACT) {
		printf("ファイルを展開しました\n");
	} else if (mode == MODE_TEST) {
		printf("チェックOK\n");
	}
#endif	/* REPLACE_VIS */

	return 0;
}

static void usage(void)
{
#ifdef REPLACE_VIS
	printf("使用方法: replace_vis [オプション] VISファイル 出力VISファイル ディレクトリ ...\n");
	printf("機    能: VISアーカイブ内のファイルを入れ替えます\n");
#else	/* REPLACE_VIS */
	printf("使用方法: extract_vis [オプション] VISファイル [出力ディレクトリ]\n");
	printf("機    能: VISアーカイブを展開します\n");
#endif	/* REPLACE_VIS */
	printf("オプション:\n");
	printf("    -B|--big-endian     ビッグエンディアンを指定します\n");
	printf("    -L|--little-endian  リトルエンディアンを指定します\n");
	printf("                        ファイルテーブルの読み込みをどちらで行うか指定します\n");
	printf("                        どちらの指定もないときは自動判別します\n");
	printf("    -k|--key VISキー    マスク解除用のキーを指定します\n");
	printf("    -K|--key-file VISキーファイル\n");
	printf("                        マスク解除用のキーを指定したファイルから読み込みます\n");
	printf("                        指定のないときはプログラムと同じディレクトリにある\n");
	printf("                        %sを使用します\n", default_key_file);
#ifdef REPLACE_VIS
	printf("       --much-index-only\n");
	printf("                        置換ファイルかどうかをファイル名最初の5桁数字部分だけで\n");
	printf("                        判定します\n");
	printf("                        数字部分が4桁以下や6桁以上のファイルは対象になりません\n");
	printf("                        このオプションを指定しないときはファイル名が\n");
	printf("                        5桁数字.拡張子 のものを対象とします\n");
	printf("    -i|--index インデックス\n");
	printf("                        インデックスで指定したファイルのみ置換します\n");
	printf("                        このオプションは複数回指定可能です\n");
	printf("                        指定のないときはディレクトリ内にある対象ファイル\n");
	printf("                        すべてを置換します\n");
	printf("    -e|--ext 拡張子     指定した拡張子のファイルのみ置換します\n");
#else	/* REPLACE_VIS */
	printf("    -l|--list           ファイル内容を一覧表示します\n");
	printf("                        ファイル展開はしません\n");
	printf("    -t|--test           圧縮データが解凍できるか確認します\n");
	printf("                        ファイル展開はしません\n");
	printf("    -i|--index インデックス\n");
	printf("                        インデックスで指定したファイルのみ展開します\n");
	printf("                        このオプションは複数回指定可能です\n");
	printf("                        指定のないときはすべて展開します\n");
	printf("    -e|--ext 拡張子     指定した拡張子のファイルのみ展開します\n");
#endif	/* REPLACE_VIS */
	printf("       --verbose        冗長出力モード\n");
	printf("    -?|--help           使用方法を表示します\n");
	printf("\n");
#ifdef REPLACE_VIS
	printf("ディレクトリは複数指定可能です\n");
	printf("置き換えるファイルが複数ある場合は先に指定されたディレクトリのものを使用します\n");
	printf("置き換えるファイルの名前の大文字小文字は区別しません\n");
#else	/* REPLACE_VIS */
	printf("出力ディレクトリの指定がないときはカレントディレクトリに展開します\n");
#endif	/* REPLACE_VIS */
}

#ifdef REPLACE_VIS
static int replace_vis(const char *in_file, const char *out_file, char **replace_dirs, int num_replace_dirs, int *vis_index, int num_vis_index, const char *ext, enum byte_order_e byte_order, const char *vis_key, int opt_much_index_only)
#else	/* REPLACE_VIS */
static int extract_vis(const char *in_file, const char *out_dir, int *vis_index, int num_vis_index, const char *ext, enum byte_order_e byte_order, const char *vis_key, enum run_mode_e mode)
#endif	/* REPLACE_VIS */
{
	int result = -1;

	unsigned char *vis_data = NULL;
	long vis_size = 0;

	unsigned int num_files = 0;
	long vis_offset = 0;
	long base_offset = 0;
#ifdef REPLACE_VIS
	long new_offset = 0;
	int num_replace_files = 0;
#endif	/* REPLACE_VIS */
	struct vis_file_info_t *info = NULL;
	size_t alloc_size = 0;
#if 0
	int lame_vis = 0;
#endif
	unsigned char *dest_buf = NULL;
	unsigned long dest_size = 0;

#ifndef REPLACE_VIS
	char *name_buf = NULL;
	int name_buf_size = 0;
	int print_header = 1;
	int cur_vis_index = 0;
#endif	/* not REPLACE_VIS */

	int i;

	if (read_vis_file(in_file, &vis_data, &vis_size) < 0) {
		printf("ERROR: VISファイルを読めませんでした: %s\n", in_file);
		goto ERROR_EXIT;
	}

	vis_offset = 4;
	if (vis_size < vis_offset + 4) {
		printf("ERROR: VISファイルが小さすぎます: %s\n", in_file);
		goto ERROR_EXIT;
	}

	if (byte_order == BYTE_ORDER_BIG_ENDIAN || byte_order == BYTE_ORDER_LITTLE_ENDIAN) {
		num_files = get_int4(vis_data + vis_offset, byte_order);
	} else {
		unsigned int be = get_int4(vis_data + vis_offset, BYTE_ORDER_BIG_ENDIAN);
		unsigned int le = get_int4(vis_data + vis_offset, BYTE_ORDER_LITTLE_ENDIAN);

		if (MAX_VIS_FILES < be && MAX_VIS_FILES < le) {
			printf("ERROR: ファイル数が多すぎます: %s\n", in_file);
			goto ERROR_EXIT;
		} else if (be <= MAX_VIS_FILES && MAX_VIS_FILES < le) {
			byte_order = BYTE_ORDER_BIG_ENDIAN;
		} else if (le <= MAX_VIS_FILES && MAX_VIS_FILES < be) {
			byte_order = BYTE_ORDER_LITTLE_ENDIAN;
		} else {
			byte_order = BYTE_ORDER_BIG_ENDIAN;
		}

		if (byte_order == BYTE_ORDER_BIG_ENDIAN) {
			num_files = be;
			printf("big endian\n");
		} else {
			num_files = le;
			printf("little endian\n");
		}
	}
	printf("num files: %d\n", num_files);
	vis_offset = 8;

	if (MAX_VIS_FILES < num_files) {
		printf("ERROR: ファイル数が多すぎます: %s\n", in_file);
		goto ERROR_EXIT;
	}

	if (vis_size < (vis_offset + 3 + 16 * num_files + 3)) {
		printf("ERROR: VISファイルが小さすぎます: %s\n", in_file);
		goto ERROR_EXIT;
	}

	if (vis_key == NULL) {
		/* vis_key_table[]から使用するキーを探す */
		int i;
		unsigned char hdrend[6];
		for (i = 0; (vis_key = vis_key_table[i]) != NULL; i++) {
			memcpy(hdrend    , vis_data + vis_offset                     , 3);
			memcpy(hdrend + 3, vis_data + vis_offset + 3 + 16 * num_files, 3);
			xor(hdrend, 6, vis_key);
			if (memcmp(hdrend, "HDREND", 6) == 0) {
				break;
			}
		}
		if (vis_key == NULL) {
			printf("ERROR: VISキーが見つかりません: %s\n", in_file);
			goto ERROR_EXIT;
		} else {
			printf("vis key: %s\n", vis_key);
		}
	}

	if (0 < num_vis_index) {
		int j;
		for (j = 0; j < num_vis_index; j++) {
			if (num_files <= vis_index[j]) {
				printf("ERROR: 指定したインデックスのファイルはありません: %s: %05d\n", in_file, vis_index[j]);
				goto ERROR_EXIT;
			}
		}
	}

	xor(vis_data + vis_offset, 3 + 16 * num_files + 3, vis_key);
	vis_offset += 3;

	alloc_size = sizeof(struct vis_file_info_t) * num_files;
	if ((info = (struct vis_file_info_t *)malloc(alloc_size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		goto ERROR_EXIT;
	}
	memset(info, 0, alloc_size);

#ifndef REPLACE_VIS
	if (mode == MODE_EXTRACT) {
		if (out_dir == NULL) {
			name_buf_size = 5 + 1 + 32;	/* 00000.(ext) */
		} else {
			name_buf_size = strlen(out_dir) + 1 + 5 + 1 + 32;	/* (out_dir)/00000.(ext) */
		}
		if ((name_buf = (char *)malloc(name_buf_size)) == NULL) {
			printf("ERROR: memory exhausted!\n");
			goto ERROR_EXIT;
		}
		memset(name_buf, 0, name_buf_size);

		if (out_dir != NULL) {
			if (make_directory(out_dir, 0) < 0) {
				printf("WARNING: 出力ディレクトリを作成できません: %s\n", out_dir);
			}
		}
	}
#endif	/* not REPLACE_VIS */

	base_offset = 8 + 3 + 16 * num_files + 3;
	for (i = 0; i < num_files; i++) {
		info[i].offset = get_int4(vis_data + vis_offset     , byte_order) + base_offset;
		info[i].z_size = get_int4(vis_data + vis_offset +  4, byte_order);
		info[i].size   = get_int4(vis_data + vis_offset +  8, byte_order);
		info[i].flag   = get_int4(vis_data + vis_offset + 12, byte_order);
		vis_offset += 16;

#if 0
		if (i == 1 && 0xFFFF < info[i].flag) {
			lame_vis = 1;
		}

		if (lame_vis != 0) {
			lame_vis_fix(info, i, &lame_vis);
		}
#endif

		if (vis_size < info[i].offset + info[i].z_size) {
			printf("ERROR: 不正なオフセットまたはサイズです: %s: %05d\n", in_file, i);
			goto ERROR_EXIT;
		}

		{
			int j;
			for (j = 0; j < i; j++) {
				if (info[j].offset == info[i].offset && info[j].z_size == info[i].z_size) {
					printf("WARNING: 同じ領域を指定しているファイルがあります: %s: %05d %05d\n", in_file, j, i);
				} else if ((info[i].offset <= info[j].offset && info[j].offset < info[i].offset + info[i].z_size) ||
						   (info[j].offset <= info[i].offset && info[i].offset < info[j].offset + info[j].z_size)) {
					printf("ERROR: ファイル領域の一部が重なっています: %s: %05d %05d\n", in_file, j, i);
					goto ERROR_EXIT;
				}
			}
		}

		info[i].data = vis_data + info[i].offset;

		if (IS_XML1(info[i].flag) || IS_XML2(info[i].flag)) {
			info[i].ext = "xml";
		} else {
			/*
			 * 以下のゲームのセーブデータはXML形式でflag=0x00000002
			 *
			 * A New Beginning: Final Cut       2.0.3.0395
			 * Edna & Harvey: Harvey's New Eyes 1.3.4.390
			 * Memoria                          1.2.3.0341
			 * The Dark Eye: Chains of Satinav  1.0
			 * The Night of The Rabbit          1.2.3.0389
			 */
			if (IS_VIS_KEY_MASKED(info[i].flag)) {
				/* マスク解除 */
				xor(info[i].data, info[i].z_size, vis_key);
			}

			get_ext(&(info[i]));
		}

#ifndef REPLACE_VIS
		if (0 < num_vis_index) {
			/* 指定したインデックスのみ */
			int found = 0;
			while (cur_vis_index < num_vis_index && vis_index[cur_vis_index] <= i) {
				if (vis_index[cur_vis_index++] == i) {
					found = 1;
					break;
				}
			}
			if (found == 0) {
				continue;
			}
		}

		if (ext != NULL) {
			/* 指定した拡張子のみ */
			if (strcasecmp(info[i].ext, ext) != 0) {
				continue;
			}
		}

		if (print_header) {
			print_header = 0;
			printf("offset   z-size   size     flag     name\n");
			printf("-------- -------- -------- -------- ----------------\n");
		}
		printf("%08lX %08lX %08lX %08X %05d.%s\n", info[i].offset, info[i].z_size, info[i].size, info[i].flag, i, info[i].ext);

		if (mode == MODE_LIST) {
			continue;
		}

		if (out_dir == NULL) {
			snprintf(name_buf, name_buf_size, "%05d.%s", i, info[i].ext);
		} else {
			snprintf(name_buf, name_buf_size, "%s%c%05d.%s", out_dir, PATH_SEP, i, info[i].ext);
		}

#if 0
		if (info[i].size != info[i].z_size) {
			if (! IS_XML1(info[i].flag)) {
				info[i].offset += 8;
				info[i].z_size -= 8;
			}
		}
#endif

		if (IS_XML1(info[i].flag)) {
			if (IS_VIS_KEY_MASKED(info[i].flag)) {
				/* マスク解除 */
				xor(info[i].data, info[i].z_size, vis_key);
			}

			dest_size = info[i].size;
			if ((dest_buf = (unsigned char *)malloc(dest_size)) == NULL) {
				printf("ERROR: memory exhausted!\n");
				goto ERROR_EXIT;
			}

			if (uncompress(dest_buf, &dest_size, info[i].data, info[i].z_size) != Z_OK) {
				printf("ERROR: 解凍できませんでした: %s: %05d\n", in_file, i);
				goto ERROR_EXIT;
			}

			if (dest_size != info[i].size) {
				printf("ERROR: 解凍後のサイズがヘッダのサイズと一致しません: %s: %05d header=%08lX uncompressed=%08lX\n", in_file, i, info[i].size, dest_size);
				goto ERROR_EXIT;
			}
		} else if (IS_XML2(info[i].flag)) {
			unsigned char *src_buf = info[i].data;
			unsigned int src_size = info[i].z_size;
			unsigned char *dest_buf2 = NULL;
			unsigned long dest_size2 = 0;			/* 解凍済みサイズ */
			int j = 0;

			dest_size = info[i].size;
			if ((dest_buf = (unsigned char *)malloc(dest_size)) == NULL) {
				printf("ERROR: memory exhausted!\n");
				goto ERROR_EXIT;
			}

			dest_buf2 = dest_buf;
			while (1) {
				unsigned int size = 0;
				unsigned int z_size = 0;
				unsigned long work_size = dest_size - dest_size2;

				if (src_size < 8) {
					printf("ERROR: 解凍できません: %s: %05d\n", in_file, i);
					goto ERROR_EXIT;
				}
				size   = get_int4(src_buf    , byte_order);
				z_size = get_int4(src_buf + 4, byte_order);
				src_buf  += 8;
				src_size -= 8;

				if (0 < verbose_level) {
					printf("block %3.1d: %08X %08X\n", j, size, z_size);
				}

				if (z_size == 0) {
					break;
				}

				if (src_size < z_size) {
					printf("ERROR: 解凍できません: %s: %05d\n", in_file, i);
					goto ERROR_EXIT;
				}

				if (IS_VIS_KEY_MASKED(info[i].flag)) {
					/* マスク解除 */
					xor(src_buf, z_size, vis_key);
				}

				if (uncompress(dest_buf2, &work_size, src_buf, z_size) != Z_OK) {
					printf("ERROR: 解凍できません: %s: %05d\n", in_file, i);
					goto ERROR_EXIT;
				}

				if (work_size != size) {
					printf("ERROR: 解凍後のサイズがブロックヘッダのサイズと一致しません: %s: %05d [%d] header=%08X uncompressed=%08lX\n", in_file, i, j, size, work_size);
					goto ERROR_EXIT;
				}

				src_buf    += z_size;
				src_size   -= z_size;
				dest_buf2  += work_size;
				dest_size2 += work_size;
				j++;
			}

			if (info[i].size != dest_size2) {
				printf("ERROR: 解凍後のサイズがヘッダのサイズと一致しません: %s: %05d header=%08lX uncompressed=%08lX\n", in_file, i, info[i].size, dest_size2);
				goto ERROR_EXIT;
			}
		}

		if (dest_buf != NULL) {
			if (mode == MODE_EXTRACT && write_file(name_buf, dest_buf, info[i].size) < 0) {
				printf("ERROR: ファイル作成できませんでした: %s\n", name_buf);
				goto ERROR_EXIT;
			}
		} else if (strcmp(info[i].ext, "vbin") == 0) {
			unsigned char *src_buf = info[i].data;
			unsigned long src_size = info[i].z_size;
			unsigned int size = 0;
			unsigned int z_size = 0;

			if (info[i].z_size <= 16) {
				printf("ERROR: 解凍できません: %s: %05d\n", in_file, i);
				goto ERROR_EXIT;
			}

			size   = get_int4(src_buf +  8, BYTE_ORDER_LITTLE_ENDIAN);
			z_size = get_int4(src_buf + 12, BYTE_ORDER_LITTLE_ENDIAN);
			src_buf  += 16;
			src_size -= 16;

			if (0 < verbose_level) {
				printf("VBIN: original=%08X compressed=%08X\n", size, z_size);
			}

			if (z_size < size) {
				if ((dest_buf = (unsigned char *)malloc(size)) == NULL) {
					printf("ERROR: memory exhausted!\n");
					goto ERROR_EXIT;
				}

				dest_size = size;
				if (uncompress(dest_buf, &dest_size, src_buf, src_size) != Z_OK) {
					printf("ERROR: 解凍できません: %s: %05d\n", in_file, i);
					goto ERROR_EXIT;
				}

				if (dest_size != size) {
					printf("ERROR: 解凍後のサイズがヘッダのサイズと一致しません: %s: %05d header=%08X uncompressed=%08lX\n", in_file, i, size, dest_size);
					goto ERROR_EXIT;
				}

				if (mode == MODE_EXTRACT && write_file(name_buf, dest_buf, dest_size) < 0) {
					printf("ERROR: ファイル作成できませんでした: %s\n", name_buf);
					goto ERROR_EXIT;
				}
			} else {
				if (mode == MODE_EXTRACT && write_file(name_buf, info[i].data + 16, info[i].size - 16) < 0) {
					printf("ERROR: ファイル作成できませんでした: %s\n", name_buf);
					goto ERROR_EXIT;
				}
			}
		} else {
			if (mode == MODE_EXTRACT && write_file(name_buf, info[i].data, info[i].size) < 0) {
				printf("ERROR: ファイル作成できませんでした: %s\n", name_buf);
				goto ERROR_EXIT;
			}
		}

		if (dest_buf != NULL) {
			free(dest_buf);
			dest_buf = NULL;
		}
#endif	/* not REPLACE_VIS */
	}

#ifdef REPLACE_VIS
	for (i = 0; i < num_replace_dirs; i++) {
		int result;
		if ((result = read_replace_files(info, num_files, replace_dirs[i], vis_index, num_vis_index, ext, opt_much_index_only)) < 0) {
			printf("ERROR: 置換対象ファイルを読み込めませんでした: %s\n", replace_dirs[i]);
			goto ERROR_EXIT;
		}
		num_replace_files += result;
	}

	if (num_replace_files <= 0) {
		printf("ERROR: 置換するファイルがありません\n");
		goto ERROR_EXIT;
	}

	/* 再構成 */
	vis_offset = 8 + 3;
	for (i = 0; i < num_files; i++, vis_offset += 16) {
		if (info[i].new_data == NULL) {
			/* 置換していないファイル */
			info[i].offset = new_offset;
			new_offset += info[i].z_size;

			if (!IS_XML1(info[i].flag) && !IS_XML2(info[i].flag) && IS_VIS_KEY_MASKED(info[i].flag)) {
				/* マスク */
				xor(info[i].data, info[i].z_size, vis_key);
			}

			put_int4(vis_data + vis_offset     , info[i].offset, byte_order);
			put_int4(vis_data + vis_offset +  4, info[i].z_size, byte_order);
			put_int4(vis_data + vis_offset +  8, info[i].size  , byte_order);
			/* フラグは変更する必要なし
			put_int4(vis_data + vis_offset + 12, info[i].flag  , byte_order); */
			continue;
		}

		/* 置換したファイル */
		if (IS_XML1(info[i].flag)) {
			/* 圧縮 */
			if (0 < verbose_level) {
				printf("compress XML\n");
			}

			dest_size = info[i].new_size;
			if ((dest_buf = (unsigned char *)malloc(dest_size)) == NULL) {
				printf("ERROR: memory exhausted!\n");
				goto ERROR_EXIT;
			}

			if (compress(dest_buf, &dest_size, info[i].new_data, info[i].new_size) != Z_OK) {
				printf("ERROR: 圧縮できませんでした: %05d\n", i);
				goto ERROR_EXIT;
			}

			/* 圧縮後のサイズに縮小 */
			if ((dest_buf = (unsigned char *)realloc(dest_buf, dest_size)) == NULL) {
				printf("ERROR: memory exhausted!\n");
				goto ERROR_EXIT;
			}

			info[i].new_z_size = dest_size;

			free(info[i].new_data);
			info[i].new_data = dest_buf;
			dest_buf = NULL;

			if (0 < verbose_level) {
				printf("XML: original=%08lX compressed=%08lX\n", info[i].new_size, info[i].new_z_size);
			}

			if (IS_VIS_KEY_MASKED(info[i].flag)) {
				/* マスク */
				xor(info[i].new_data, info[i].new_z_size, vis_key);
			}
		} else if (IS_XML2(info[i].flag)) {
			/* 圧縮 */
			/*
			 * 以下のゲームでは0x00200000バイト単位で圧縮
			 * 最後は0xFFEEFFEE,0x00000000
			 *
			 * A New Beginning: Final Cut       2.0.3.0395
			 * Chaos on Deponia                 2.0.3.2299
			 * Deponia                          2.0.3.1281
			 * Edna & Harvey: Harvey's New Eyes 1.3.4.390
			 * Memoria                          1.2.3.0341
			 * The Dark Eye: Chains of Satinav  1.0
			 * The Night of The Rabbit          1.2.3.0389
			 */
#if 0
			unsigned int block_size = 0;
			unsigned int end_mark   = 0;
#else
			const unsigned int block_size = 0x00200000;
			const unsigned int end_mark   = 0xFFEEFFEE;
#endif
			const unsigned char *src_buf = info[i].new_data;
			unsigned int src_size = info[i].new_size;
			unsigned char *dest_buf2 = NULL;
			unsigned int dest_size2 = 0;			/* 圧縮済みサイズ */
			int j = 0;

			if (0 < verbose_level) {
				printf("compress XML\n");
			}

#if 0
			/* 変更前データからブロックサイズとブロック終端マークを取得する */
			if (get_block_size_and_end_mark(&(info[i]), &block_size, &end_mark, byte_order) < 0) {
				printf("ERROR: ブロックサイズおよびブロック終端マークを取得できません: %s: %05d\n", in_file, i);
				goto ERROR_EXIT;
			}
#endif

			dest_size = info[i].new_size;
			if ((dest_buf = (unsigned char *)malloc(dest_size)) == NULL) {
				printf("ERROR: memory exhausted!\n");
				goto ERROR_EXIT;
			}

			dest_buf2 = dest_buf;
			while (0 < src_size) {
				unsigned int size = block_size;
				unsigned long z_size = dest_size - dest_size2 - 8;	/* ブロックヘッダの8バイト分少なくする */

				if (z_size <= 0) {
					printf("ERROR: 圧縮できません: %05d\n", i);
					goto ERROR_EXIT;
				}

				/* ブロックサイズに分割して圧縮 */
				if (src_size < block_size) {
					size = src_size;
				}
				if (compress(dest_buf2 + 8, &z_size, src_buf, size) != Z_OK) {
					printf("ERROR: 圧縮できません: %05d\n", i);
					goto ERROR_EXIT;
				}

				put_int4(dest_buf2    , size  , byte_order);
				put_int4(dest_buf2 + 4, z_size, byte_order);
				dest_buf2  += 8;
				dest_size2 += 8;

				if (0 < verbose_level) {
					printf("block %3.1d: %08X %08lX\n", j, size, z_size);
				}

				if (IS_VIS_KEY_MASKED(info[i].flag)) {
					/* マスク */
					xor(dest_buf2, z_size, vis_key);
				}

				src_buf    += size;
				src_size   -= size;
				dest_buf2  += z_size;
				dest_size2 += z_size;
				j++;
			}

			if (dest_size - dest_size2 < 8) {
				printf("ERROR: 圧縮できません: %05d\n", i);
				goto ERROR_EXIT;
			}

			put_int4(dest_buf2    , end_mark, byte_order);
			put_int4(dest_buf2 + 4, 0       , byte_order);
			dest_buf2  += 8;
			dest_size2 += 8;

			if (0 < verbose_level) {
				printf("block %3.1d: %08X 00000000\n", j, end_mark);
			}

			/* 圧縮後のサイズに縮小 */
			if ((dest_buf = (unsigned char *)realloc(dest_buf, dest_size2)) == NULL) {
				printf("ERROR: memory exhausted!\n");
				goto ERROR_EXIT;
			}

			info[i].new_z_size = dest_size2;

			free(info[i].new_data);
			info[i].new_data = dest_buf;
			dest_buf = NULL;

			if (0 < verbose_level) {
				printf("XML: original=%08lX compressed=%08lX\n", info[i].new_size, info[i].new_z_size);
			}
		} else if (strcmp(info[i].ext, "vbin") == 0) {
			/* 圧縮 */
			if (0 < verbose_level) {
				printf("compress VBIN\n");
			}

			/* VBINヘッダの16バイトを追加して確保 */
			dest_size = info[i].new_size;
			if ((dest_buf = (unsigned char *)malloc(dest_size + 16)) == NULL) {
				printf("ERROR: memory exhausted!\n");
				goto ERROR_EXIT;
			}

			if (compress(dest_buf + 16, &dest_size, info[i].new_data, info[i].new_size) != Z_OK) {
				printf("ERROR: 圧縮できませんでした: %05d\n", i);
				goto ERROR_EXIT;
			}

			/* VBINヘッダ作成 */
			memcpy(dest_buf, "VBIN", 4);
			memset(dest_buf + 4, 0, 4);			/* 不明 - 0でも問題なくゲームで使用できている */
			put_int4(dest_buf +  8, info[i].new_size, BYTE_ORDER_LITTLE_ENDIAN);
			put_int4(dest_buf + 12, dest_size       , BYTE_ORDER_LITTLE_ENDIAN);

			if (0 < verbose_level) {
				printf("VBIN: original=%08lX compressed=%08lX\n", info[i].new_size, dest_size);
			}

			dest_size += 16;					/* ヘッダの16バイトを追加 */

			/* 圧縮後のサイズに縮小 */
			if ((dest_buf = (unsigned char *)realloc(dest_buf, dest_size)) == NULL) {
				printf("ERROR: memory exhausted!\n");
				goto ERROR_EXIT;
			}

			info[i].new_z_size = dest_size;
			info[i].new_size   = dest_size;

			free(info[i].new_data);
			info[i].new_data = dest_buf;
			dest_buf = NULL;
		} else {
			/* 圧縮しない */
			info[i].new_z_size = info[i].new_size;

			if (IS_VIS_KEY_MASKED(info[i].flag)) {
				/* マスク */
				xor(info[i].new_data, info[i].new_z_size, vis_key);
			}
		}

		info[i].offset = new_offset;
		new_offset += info[i].new_z_size;

		put_int4(vis_data + vis_offset     , info[i].offset    , byte_order);
		put_int4(vis_data + vis_offset +  4, info[i].new_z_size, byte_order);
		put_int4(vis_data + vis_offset +  8, info[i].new_size  , byte_order);
		/* フラグは変更する必要なし
		put_int4(vis_data + vis_offset + 12, info[i].flag      , byte_order); */
	}

	/* ヘッダをマスク */
	xor(vis_data + 8, 3 + 16 * num_files + 3, vis_key);

	/* VISファイル作成 */
	if (write_vis_file(out_file, vis_data, info, num_files) < 0) {
		printf("ERROR: VISファイルを作成できませんでした: %s\n", out_file);
		goto ERROR_EXIT;
	}
#endif	/* REPLACE_VIS */

	result = 0;

ERROR_EXIT:
	if (info != NULL) {
#ifdef REPLACE_VIS
		for (i = 0; i < num_files; i++) {
			if (info[i].new_data != NULL) {
				free(info[i].new_data);
			}
		}
#endif	/* REPLACE_VIS */
		free(info);
	}
	if (vis_data != NULL) {
		free(vis_data);
	}
	if (dest_buf != NULL) {
		free(dest_buf);
	}

	return result;
}

#ifdef REPLACE_VIS
#if 0
static int get_block_size_and_end_mark(const struct vis_file_info_t *info, unsigned int *block_size, unsigned int *end_mark, enum byte_order_e byte_order)
{
	int result = -1;
	unsigned int src_size = info->z_size;
	const unsigned char *src = info->data;
	int found_block_size = 0;
	int found_end_mark = 0;

	while (1) {
		unsigned int size = 0;
		unsigned int z_size = 0;

		if (src_size < 8) {
			printf("ERROR: ブロックサイズを取得できません\n");
			goto ERROR_EXIT;
		}

		size   = get_int4(src    , byte_order);
		z_size = get_int4(src + 4, byte_order);
		src      += 8;
		src_size -= 8;

		if (z_size == 0) {
			*end_mark = size;
			found_end_mark = 1;
			if (0 < verbose_level) {
				printf("found end mark  : %08X\n", *end_mark);
			}
			break;
		}

		if (src_size < z_size) {
			printf("ERROR: ブロックサイズを取得できません\n");
			goto ERROR_EXIT;
		}

		if (found_block_size == 0) {
			*block_size = size;
			found_block_size = 1;
			if (0 < verbose_level) {
				printf("found block size: %08X\n", *block_size);
			}
		} else if (src_size != (z_size + 8) && *block_size != size && found_block_size == 1) {
			printf("WARNING: ブロックサイズが一定ではありません\n");
			found_block_size = 2;
		}

		if (src_size < z_size) {
			printf("ERROR: ブロックサイズを取得できません\n");
			goto ERROR_EXIT;
		}

		src      += z_size;
		src_size -= z_size;
	}

	if (found_block_size != 0 && found_end_mark != 0) {
		result = 0;
	}

ERROR_EXIT:

	return result;
}
#endif
#endif	/* REPLACE_VIS */

static int check_vis_key(char *des, const char *src)
{
	int result = -1;
	int num_upper = 0;
	char *save = des;
	int i;

	if (strlen(src) != VIS_KEY_SIZE) {
		printf("ERROR: 不正なVISキーです: %s\n", src);
		goto ERROR_EXIT;
	}

	for (i = 0; i < VIS_KEY_SIZE; i++) {
		if (!isxdigit((unsigned char)src[i])) {
			printf("ERROR: 不正なVISキーです: %s\n", src);
			goto ERROR_EXIT;
		}
		if (isupper(src[i])) {
			num_upper++;
			des[i] = tolower(src[i]);
		} else {
			des[i] = src[i];
		}
	}
	des[i] = 0;

	if (num_upper != 0) {
		printf("WARNING: VISキーに含まれる英大文字を英小文字に変換しました: %s\n", save);
	}

	result = 0;

ERROR_EXIT:
	return result;
}

static int read_key_file(const char *filename)
{
	int result = -1;
	FILE *fp = NULL;
	int line = 0;
	int num_keys = 0;
	char vis_key_buf[VIS_KEY_SIZE + 1];

	int read_status;
	char *buffer = NULL;
	size_t buffer_size = 0;

	if (0 < verbose_level) {
		printf("key file: %s\n", filename);
	}

	if ((fp = fopen(filename, "rt")) == NULL) {
		printf("ERROR: %d: ファイルを開けません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	while (1) {
		char *ptr = NULL;

		if ((read_status = read_next_line(&buffer, &buffer_size, fp)) < 0) {
			printf("ERROR: ファイルを読めません: %s line %d\n", filename, line);
			goto ERROR_EXIT;
		} else if (read_status == 0) {
			break;
		}

		line++;
		ptr = buffer;

		if (line == 1) {
			ptr = skip_bom(ptr);
		}

		ptr = strtok(ptr, " \t\r\n");
		if (ptr == NULL || !isxdigit(*(unsigned char *)ptr)) {
			continue;
		}

		if (check_vis_key(vis_key_buf, ptr) < 0) {
			printf("ERROR: 不正なVISキーです: %s\n", ptr);
			goto ERROR_EXIT;
		}

		if (num_keys == 0) {
			vis_key_table = (char **)malloc(sizeof(char *) * 2);
		} else {
			vis_key_table = (char **)realloc(vis_key_table, sizeof(char *) * (num_keys + 2));
		}
		if (vis_key_table == NULL) {
			printf("ERROR: memory exhausted!\n");
			goto ERROR_EXIT;
		}

		if ((vis_key_table[num_keys] = strdup(vis_key_buf)) == NULL) {
			printf("ERROR: memory exhausted!\n");
			goto ERROR_EXIT;
		}

		if (0 < verbose_level) {
			printf("vis key %3.1d: %s\n", num_keys, vis_key_table[num_keys]);
		}

		num_keys++;
		vis_key_table[num_keys] = NULL;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (buffer != NULL) {
		free(buffer);
	}

	return result;
}

static int read_vis_file(const char *filename, unsigned char **data, long *size)
{
	int result = -1;
	struct stat st;
	FILE *fp = NULL;
	const char *vis_sig = "VIS3";
	char buffer[4];

	*data = NULL;
	*size = 0;

	if (stat(filename, &st) < 0) {
		printf("ERROR: %d: ファイル情報を取得できません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}
	*size = st.st_size;

	if ((fp = fopen(filename, "rb")) == NULL) {
		printf("ERROR: %d: ファイルを開けません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	if (fread(buffer, 4, 1, fp) != 1) {
		printf("ERROR: %d: ファイルを読めません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	if (memcmp(buffer, vis_sig, 4) != 0) {
		printf("ERROR: 未対応のファイルです: %s\n", filename);
		goto ERROR_EXIT;
	}

	if ((*data = (unsigned char *)malloc(*size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		goto ERROR_EXIT;
	}

	memcpy(*data, buffer, 4);
	if (fread((*data) + 4, *size - 4, 1, fp) != 1) {
		printf("ERROR: %d: ファイルを読めません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (result < 0) {
		if (*data != NULL) {
			free(*data);
		}
		*data = NULL;
		*size = 0;
	}

	return result;
}

#ifdef REPLACE_VIS
static int read_replace_files(struct vis_file_info_t *info, unsigned int num_files, const char *replace_dir, int *vis_index, int num_vis_index, const char *ext, int opt_much_index_only)
{
	int result = -1;
	DIR *dirp = NULL;
	struct dirent *dp = NULL;
	char *name_buf = NULL;
	size_t name_buf_size = 0;
	size_t dirname_len = strlen(replace_dir);
	int num_replace_files = 0;

	if ((dirp = opendir(replace_dir)) == NULL) {
		printf("ERROR: ディレクトリを開けません: %s\n", replace_dir);
		goto ERROR_EXIT;
	}

	while ((dp = readdir(dirp)) != NULL) {
		int len;
		int index;
		size_t name_size;

		for (len = 0; len < 6 && isdigit((unsigned char)(dp->d_name)[len]); len++) ;
		if (len != 5) {
			/* 数字が5桁でない */
			continue;
		}

		index = atoi(dp->d_name);

		if (num_files <= index) {
			/* 範囲外 */
			continue;
		}

		if (info[index].new_data != NULL) {
			/* 既に置換対象ファイルを見つけて読み込み済み */
			continue;
		}

		if (0 < num_vis_index) {
			/* 指定したインデックスのみ */
			int found = 0;
			int i;
			for (i = 0; i < num_vis_index; i++) {
				if (vis_index[i] == index) {
					found = 1;
					break;
				}
			}
			if (found == 0) {
				continue;
			}
		}

		if (ext != NULL) {
			/* 指定した拡張子のみ */
			/* 置換対象ファイルのもつ拡張子ではなくアーカイブ内のファイルの拡張子を指定する */
			if (strcasecmp(info[index].ext, ext) != 0) {
				continue;
			}
		}

		if (opt_much_index_only == 0) {
			if (dp->d_name[5] != '.' || strcasecmp(dp->d_name + 6, info[index].ext) != 0) {
				/* 拡張子が一致しない */
				continue;
			}
		}

		name_size = dirname_len + 1 + strlen(dp->d_name) + 1;
		if (name_buf_size == 0) {
			name_buf = (char *)malloc(name_size);
			name_buf_size = name_size;
		} else if (name_buf_size < name_size) {
			name_buf = (char *)realloc(name_buf, name_size);
			name_buf_size = name_size;
		}
		if (name_buf == NULL) {
			printf("ERROR: memory exhausted!\n");
			goto ERROR_EXIT;
		}
		snprintf(name_buf, name_buf_size, "%s%c%s", replace_dir, PATH_SEP, dp->d_name);

		printf("replace: %05d %s\n", index, name_buf);

		if (read_file(name_buf, &(info[index].new_data), &(info[index].new_size)) < 0) {
			printf("ERROR: ファイルを読みませんでした: %s\n", name_buf);
			goto ERROR_EXIT;
		}

		num_replace_files++;
	}
	if (errno != 0 && errno != ENOENT) {
		printf("ERROR: %d: ディレクトリを読めません: %s\n", errno, replace_dir);
		goto ERROR_EXIT;
	}

	result = num_replace_files;

ERROR_EXIT:
	if (dirp != NULL) {
		closedir(dirp);
	}
	if (name_buf != NULL) {
		free(name_buf);
	}

	return result;
}

static int read_file(const char *filename, unsigned char **data, long *size)
{
	int result = -1;
	struct stat st;
	FILE *fp = NULL;

	*data = NULL;
	*size = 0;

	if (stat(filename, &st) < 0) {
		printf("ERROR: %d: ファイル情報を取得できません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	/* VISアーカイブのファイルサイズは4バイト */
	if (0xFFFFFFFF < st.st_size) {
		printf("ERROR: ファイルサイズが大きすぎます: %s\n", filename);
		goto ERROR_EXIT;
	}

	*size = st.st_size;

	if ((fp = fopen(filename, "rb")) == NULL) {
		printf("ERROR: %d: ファイルを開けません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	if ((*data = (unsigned char *)malloc(*size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		goto ERROR_EXIT;
	}

	if (fread((*data), *size, 1, fp) != 1) {
		printf("ERROR: %d: ファイルを読めません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (result < 0) {
		if (*data != NULL) {
			free(*data);
		}
		*data = NULL;
		*size = 0;
	}

	return result;
}

static int write_vis_file(const char *filename, const unsigned char *vis_data, const struct vis_file_info_t *info, unsigned int num_files)
{
	int result = -1;
	FILE *fp = NULL;
	int i;

	if ((fp = fopen(filename, "wb")) == NULL) {
		printf("ERROR: %d: ファイルを開けません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	/* ヘッダ */
	if (fwrite(vis_data, 8 + 3 + 16 * num_files + 3, 1, fp) != 1) {
		printf("ERROR: %d: ファイルを書けません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	/* データ */
	for (i = 0; i < num_files; i++) {
		const unsigned char *data;
		size_t size;
		if (info[i].new_data == NULL) {
			data = info[i].data;
			size = info[i].z_size;
		} else {
			data = info[i].new_data;
			size = info[i].new_z_size;
		}
		if (fwrite(data, size, 1, fp) != 1) {
			printf("ERROR: %d: ファイルを書けません: %s\n", errno, filename);
			goto ERROR_EXIT;
		}
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}
#endif	/* REPLACE_VIS */

#ifndef REPLACE_VIS
static int write_file(const char *filename, const unsigned char *data, long size)
{
	int result = -1;
	FILE *fp = NULL;

	if ((fp = fopen(filename, "wb")) == NULL) {
		printf("ERROR: %d: ファイルを開けません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	if (fwrite(data, size, 1, fp) != 1) {
		printf("ERROR: %d: ファイルを書けません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}
#endif	/* not REPLACE_VIS */

static unsigned int get_int4(const unsigned char *data, enum byte_order_e byte_order)
{
	if (byte_order == BYTE_ORDER_LITTLE_ENDIAN) {
		return ((unsigned int)*data) | (((unsigned int)*(data + 1)) << 8) | (((unsigned int)*(data + 2)) << 16) | (((unsigned int)*(data + 3)) << 24);
	}
	return (((unsigned int)*data) << 24) | (((unsigned int)*(data + 1)) << 16) | (((unsigned int)*(data + 2)) << 8) | ((unsigned int)*(data + 3));
}

#ifdef REPLACE_VIS
static void put_int4(unsigned char *data, unsigned int value, enum byte_order_e byte_order)
{
	if (byte_order == BYTE_ORDER_LITTLE_ENDIAN) {
		*data++ = (unsigned char)( value        & 0xFF);
		*data++ = (unsigned char)((value >>  8) & 0xFF);
		*data++ = (unsigned char)((value >> 16) & 0xFF);
		*data++ = (unsigned char)((value >> 24) & 0xFF);
	} else {
		*data++ = (unsigned char)((value >> 24) & 0xFF);
		*data++ = (unsigned char)((value >> 16) & 0xFF);
		*data++ = (unsigned char)((value >>  8) & 0xFF);
		*data++ = (unsigned char)( value        & 0xFF);
	}
}
#endif	/* REPLACE_VIS */

static void xor(unsigned char *data, long size, const char *key)
{
	while (VIS_KEY_SIZE <= size) {
		const char *ptr = key;
		int len = VIS_KEY_SIZE;

		while (0 < len--) {
			*data++ ^= *ptr++;
		}

		size -= VIS_KEY_SIZE;
	}

	while (0 < size--) {
		*data++ ^= *key++;
	}
}

#if 0
static void lame_vis_fix(struct vis_file_info_t *info, int index, int *lame_vis)
{
	static long old_offset = 0;
	static long old_z_size = 0;

	if (index <= 4) {
		info[index].flag = 0;
	} else if (index <= 8) {
		info[index].offset = old_offset + old_z_size;
	} else {
		info[index].z_size &= 0x00FFFFFF;
		*lame_vis = 0;
	}
	old_offset = info[index].offset;
	old_z_size = info[index].z_size;
}
#endif

static void get_ext(struct vis_file_info_t *info)
{
	info->ext = "dat";

	if (info->z_size < 4) {
		return;
	}

	if (memcmp(info->data, "OggS", 4) == 0) {
		if (36 <= info->z_size && memcmp(info->data + 28, "OpusHead", 8) == 0) {
			info->ext = "opus";
		} else {
			info->ext = "ogg";
		}
	} else if (memcmp(info->data, "\x89PNG", 4) == 0) {
		info->ext = "png";
	} else if (12 <= info->z_size &&
			   memcmp(info->data    , "RIFF", 4) == 0 &&
			   memcmp(info->data + 8, "WEBP", 4) == 0) {
		info->ext = "webp";
	} else if (memcmp(info->data, "VBIN", 4) == 0) {
		info->ext = "vbin";
	} else if (6 <= info->z_size && memcmp(info->data, "<?xml ", 6) == 0) {
		info->ext = "xml";
	}
}

static int compare_index(const void *a, const void *b)
{
	unsigned int ia = *(const unsigned int *)a;
	unsigned int ib = *(const unsigned int *)b;

	if (ia < ib) {
		return -1;
	} else if (ia > ib) {
		return 1;
	}
	return 0;
}

/* EOF */
