#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "utils.h"
#include "file_utils.h"
#ifdef WIN32
#include "my_printf.h"
#define	printf	my_printf
#endif


struct font_t141_t {
	long line;

	char *t12;		/* フォント名 */
	int t13;		/* フォントID */
	char *t14;
	char *t55;
	char *t703;		/* 文字間隔補正 */
	char *t635;
	char *t249;		/* 文字セット */
	char *t250;
	char *t540;
	char *t634;
	char *t623;
	struct font_t143_t {
		int size;
		int used;
		struct font_t22_t {
			int t3;		/* 上端Y座標 */
			int t4;		/* 左端X座標 */
			int t5;		/* 下端Y座標 */
			int t6;		/* 右端X座標 */
		} *t22;
	} t143;
	struct font_t248_t {
		char *t12;
		char *t32;
		char *t7;		/* パス */
		char *t50;
		struct font_t8_t {
			char *t1;
			char *t2;
		} t8;
	} t248;
};


static void usage(void);
static int font_info(const char *in_file, const char *out_file, int output_extra_font_info);
static char *get_string(const char *src, int name_id);
static int get_integer(const char *src, int name_id);
static int write_font_info(FILE *fp, const struct font_t141_t *font, int output_extra_font_info);
static int print_safe_text(const char *text);
static int write_safe_text(FILE *fp, const char *text);
static void free_font_info(struct font_t141_t *font);
static void safe_free(void *ptr);


int main(int argc, char **argv)
{
	int num_cmd_params = 2;
	const char *xml_file = NULL;
	const char *font_file = NULL;
	int output_extra_font_info = 0;

	while (*++argv) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "--extra-font-info") == 0) {
				output_extra_font_info = 1;
			} else {
				printf("ERROR: 不正なオプションです: %s\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("ERROR: コマンドラインパラメータが多すぎます: %s\n", *argv);
				return 1;
			}
			if (num_cmd_params == 2) {
				xml_file = *argv;
			} else {
				font_file = *argv;
			}
			num_cmd_params--;
		}
	}

	if (xml_file == NULL || font_file == NULL) {
		usage();
		return 1;
	}

	if (font_info(xml_file, font_file, output_extra_font_info) < 0) {
		printf("ERROR: フォント情報を抽出できません\n");
		return 1;
	}

	return 0;
}

static void usage(void)
{
	printf("使用方法: font_info [オプション] XMLファイル 出力ファイル\n");
	printf("機    能: XMLファイルのフォント情報を保存します\n");
	printf("オプション:\n");
	printf("       --extra-font-info    追加のフォント情報を出力します\n");
	printf("    -?|--help               使用方法を表示して終了します\n");
}

static int font_info(const char *in_file, const char *out_file, int output_extra_font_info)
{
	int result = -1;
	FILE *in_fp = NULL;
	FILE *out_fp = NULL;
	long line = 0;
	char *ptr;
	struct font_t141_t font;

	int read_status;
	char *buffer = NULL;
	size_t buffer_size = 0;

	memset(&font, 0, sizeof(font));

	if ((in_fp = fopen(in_file, "rb")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, in_file);
		goto ERROR_EXIT;
	}

	if ((out_fp = fopen(out_file, "wb")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, out_file);
		goto ERROR_EXIT;
	}

	while (1) {
		if ((read_status = read_next_line(&buffer, &buffer_size, in_fp)) < 0) {
			printf("ERROR: 読み込みエラー: %s\n", in_file);
			goto ERROR_EXIT;
		} else if (read_status == 0) {
			break;
		}

		line++;

		if ((ptr = strstr(buffer, "<T141 ")) == NULL) {
			/* フォントでない */
			continue;
		}

		ptr += 5;

		font.line = line;

		if ((font.t12 = get_string(ptr, 12)) == NULL) {
			printf("ERROR: フォント名が見つかりません: %ld行目\n", line);
			goto ERROR_EXIT;
		}
		if ((font.t13 = get_integer(ptr, 13)) < 0) {
			printf("ERROR: フォントIDが見つかりません: %ld行目\n", line);
			goto ERROR_EXIT;
		}
		font.t14 = get_string(ptr, 14);
		font.t55 = get_string(ptr, 55);
		if ((font.t703 = get_string(ptr, 703)) == NULL) {
			printf("ERROR: 文字間隔が見つかりません: %ld行目\n", line);
			goto ERROR_EXIT;
		}
		if (unescape_text(font.t703, NULL) < 0) {
			printf("ERROR: 文字間隔に不正な文字があります: %ld行目\n", line);
			goto ERROR_EXIT;
		}
		font.t635 = get_string(ptr, 635);
		if ((font.t249 = get_string(ptr, 249)) == NULL) {
			printf("ERROR: 文字セットが見つかりません: %ld行目\n", line);
			goto ERROR_EXIT;
		}
		if (unescape_text(font.t249, NULL) < 0) {
			printf("ERROR: 文字セットに不正な文字があります: %ld行目\n", line);
			goto ERROR_EXIT;
		}
		font.t250 = get_string(ptr, 250);
		font.t540 = get_string(ptr, 540);
		font.t634 = get_string(ptr, 634);
		font.t623 = get_string(ptr, 623);

		while (1) {
			if ((read_status = read_next_line(&buffer, &buffer_size, in_fp)) < 0) {
				printf("ERROR: 読み込みエラー: %s\n", in_file);
				goto ERROR_EXIT;
			} else if (read_status == 0) {
				break;
			}

			line++;

			if ((ptr = strstr(buffer, "</T141>")) != NULL) {
				break;
			} else if ((ptr = strstr(buffer, "<T143>")) != NULL) {
				while (1) {
					if ((read_status = read_next_line(&buffer, &buffer_size, in_fp)) < 0) {
						printf("ERROR: 読み込みエラー: %s\n", in_file);
						goto ERROR_EXIT;
					} else if (read_status == 0) {
						break;
					}

					line++;

					if ((ptr = strstr(buffer, "</T143>")) != NULL) {
						break;
					} else if ((ptr = strstr(buffer, "<T22 ")) != NULL) {
						ptr += 4;

						if (font.t143.size <= font.t143.used) {
							const int expand_size = 100;
							size_t alloc_size = sizeof(struct font_t22_t) * (font.t143.size + expand_size);

							if (font.t143.size == 0) {
								font.t143.t22 = (struct font_t22_t *)malloc(alloc_size);
							} else {
								font.t143.t22 = (struct font_t22_t *)realloc(font.t143.t22, alloc_size);
							}
							if (font.t143.t22 == NULL) {
								printf("ERROR: memory exhausted!\n");
								goto ERROR_EXIT;
							}
							memset(&(font.t143.t22[font.t143.size]), 0, sizeof(struct font_t22_t) * expand_size);
							font.t143.size += expand_size;
						}

						if ((font.t143.t22[font.t143.used].t3 = get_integer(ptr, 3)) < 0) {
							printf("ERROR: 上端Y座標が見つかりません: %ld行目\n", line);
							goto ERROR_EXIT;
						}
						if ((font.t143.t22[font.t143.used].t4 = get_integer(ptr, 4)) < 0) {
							printf("ERROR: 左端X座標が見つかりません: %ld行目\n", line);
							goto ERROR_EXIT;
						}
						if ((font.t143.t22[font.t143.used].t5 = get_integer(ptr, 5)) < 0) {
							printf("ERROR: 下端Y座標が見つかりません: %ld行目\n", line);
							goto ERROR_EXIT;
						}
						if ((font.t143.t22[font.t143.used].t6 = get_integer(ptr, 6)) < 0) {
							printf("ERROR: 右端X座標が見つかりません: %ld行目\n", line);
							goto ERROR_EXIT;
						}

						font.t143.used++;
					}
				}
			} else if ((ptr = strstr(buffer, "<T248 ")) != NULL) {
				ptr += 5;

				font.t248.t12 = get_string(ptr, 12);
				font.t248.t32 = get_string(ptr, 32);
				if ((font.t248.t7 = get_string(ptr, 7)) == NULL) {
					printf("ERROR: パスが見つかりません: %ld行目\n", line);
					goto ERROR_EXIT;
				}
				font.t248.t50 = get_string(ptr, 50);

				while (1) {
					if ((read_status = read_next_line(&buffer, &buffer_size, in_fp)) < 0) {
						printf("ERROR: 読み込みエラー: %s\n", in_file);
						goto ERROR_EXIT;
					} else if (read_status == 0) {
						break;
					}

					line++;

					if ((ptr = strstr(buffer, "</T248>")) != NULL) {
						break;
					} else if ((ptr = strstr(buffer, "<T8 ")) != NULL) {
						ptr += 3;

						font.t248.t8.t1 = get_string(ptr, 1);
						font.t248.t8.t2 = get_string(ptr, 2);
					}
				}
			}
		}

		if (write_font_info(out_fp, &font, output_extra_font_info) < 0) {
			printf("ERROR: 書き込みエラー: %s\n", out_file);
			goto ERROR_EXIT;
		}

		free_font_info(&font);
	}

	result = 0;

ERROR_EXIT:
	if (in_fp != NULL) {
		fclose(in_fp);
	}
	if (out_fp != NULL) {
		fclose(out_fp);
	}
	if (buffer != NULL) {
		free(buffer);
	}

	return result;
}

static char *get_string(const char *src, int name_id)
{
	char name[32];
	const char *end;
	int len;
	char *des;

	snprintf(name, sizeof(name), " T%d=\"", name_id);

	if ((src = strstr(src, name)) == NULL) {
		return NULL;
	}

	src += strlen(name);
	if ((end = strchr(src, '\"')) == NULL) {
		printf("ERROR: 文字列終端の\"がありません\n");
		return NULL;
	}

	len = end - src;
	if ((des = (char *)malloc(len + 1)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		return NULL;
	}

	strncpy(des, src, len);
	des[len] = 0;

	return des;
}

static int get_integer(const char *src, int name_id)
{
	char name[32];
	char *endptr;
	int value;

	snprintf(name, sizeof(name), " T%d=\"", name_id);

	if ((src = strstr(src, name)) == NULL) {
		return -1;
	}

	src += strlen(name);
	if ((value = strtoul(src, &endptr, 10)) < 0 ||
		endptr == NULL ||
		*endptr != '\"' ||
		endptr == src) {
		printf("ERROR: 不正な数値です\n");
		return -1;
	}

	return value;
}

static int write_font_info(FILE *fp, const struct font_t141_t *font, int output_extra_font_info)
{
	int i;
	const char *src = font->t249;

	if (fprintf(fp, "[%d]\n", font->t13) < 0) {
		return -1;
	}

	if (output_extra_font_info) {
		int min_width = 0;
		int max_width = 0;
		int min_height = 0;
		int max_height = 0;
		int i;

		for (i = 0; i < font->t143.used; i++) {
			const struct font_t22_t *t22 = &(font->t143.t22[i]);
			int width = t22->t6 - t22->t4 + 1;
			int height = t22->t5 - t22->t3 + 1;
			if (min_width == 0 || width < min_width) {
				min_width = width;
			}
			if (max_width == 0 || max_width < width) {
				max_width = width;
			}
			if (min_height == 0 || height < min_height) {
				min_height = height;
			}
			if (max_height == 0 || max_height < height) {
				max_height = height;
			}
		}

		if (fprintf(fp, "#\tT12=") < 0 ||
			write_safe_text(fp, font->t12) < 0 ||
			fprintf(fp, "\n#\tT13=%d", font->t13) < 0 ||
			fprintf(fp, "\n#\tT14=") < 0 ||
			write_safe_text(fp, font->t14) < 0 ||
			fprintf(fp, "\n#\tT55=") < 0 ||
			write_safe_text(fp, font->t55) < 0 ||
			fprintf(fp, "\n#\tT703=") < 0 ||
			write_safe_text(fp, font->t703) < 0 ||
			fprintf(fp, "\n#\tT635=") < 0 ||
			write_safe_text(fp, font->t635) < 0 ||
			fprintf(fp, "\n#\tT249=") < 0 ||
			write_safe_text(fp, font->t249) < 0 ||
			fprintf(fp, "\n#\tT250=") < 0 ||
			write_safe_text(fp, font->t250) < 0 ||
			fprintf(fp, "\n#\tT540=") < 0 ||
			write_safe_text(fp, font->t540) < 0 ||
			fprintf(fp, "\n#\tT634=") < 0 ||
			write_safe_text(fp, font->t634) < 0 ||
			fprintf(fp, "\n#\tT623=") < 0 ||
			write_safe_text(fp, font->t623) < 0 ||
			fprintf(fp, "\n#\tT248\tT12=") < 0 ||
			write_safe_text(fp, font->t248.t12) < 0 ||
			fprintf(fp, "\n#\t\tT32=") < 0 ||
			write_safe_text(fp, font->t248.t32) < 0 ||
			fprintf(fp, "\n#\t\tT7=") < 0 ||
			write_safe_text(fp, font->t248.t7) < 0 ||
			fprintf(fp, "\n#\t\tT50=") < 0 ||
			write_safe_text(fp, font->t248.t50) < 0 ||
			fprintf(fp, "\n#\t\tT8\tT1=") < 0 ||
			write_safe_text(fp, font->t248.t8.t1) < 0 ||
			fprintf(fp, "\n#\t\t\tT2=") < 0 ||
			write_safe_text(fp, font->t248.t8.t2) < 0 ||
			fprintf(fp, "\n#\tmin width=%d\theight=%d\n", min_width, min_height) < 0 ||
			fprintf(fp, "#\tmax width=%d\theight=%d\n", max_width, max_height) < 0) {
			return -1;
		}
	}

	if (fprintf(fp, "name=") < 0 ||
		write_safe_text(fp, font->t12) < 0 ||
		fprintf(fp, "\npath=") < 0 ||
		write_safe_text(fp, font->t248.t7) < 0 ||
		fprintf(fp, "\nspacing=") < 0 ||
		write_safe_text(fp, font->t703) < 0 ||
		fprintf(fp, "\n") < 0) {
		return -1;
	}

	for (i = 0; i < font->t143.used; i++) {
		const struct font_t22_t *t22 = &(font->t143.t22[i]);
		unsigned long code;
		int len;

		if ((len = get_next_utf8_char(&code, src, 0)) < 0) {
			printf("ERROR: 文字セットに不正な文字が含まれています: %ld行目\n", font->line);
			return -1;
		}
		if (len == 0) {
			printf("ERROR: 文字セットが不足しています: %ld行目\n", font->line);
			return -1;
		}

		if (fprintf(fp, "char=") < 0) {
			return -1;
		}

		if (code <= 0x7F) {
			if (fprintf(fp, "%c", (char)(code & 0x7F)) < 0) {
				return -1;
			}
		} else {
			if (fwrite(src, len, 1, fp) != 1) {
				return -1;
			}
		}

		if (fprintf(fp, "\t%d\t%d\t%d\t%d\n", t22->t4, t22->t3, t22->t6, t22->t5) < 0) {
			return -1;
		}

		src += len;
	}

	if (*src != 0) {
		printf("WARNING: 座標データが不足しています: %ld行目\n", font->line);
		printf("\tname=");
		print_safe_text(font->t12);
		printf("\n\tpath=");
		print_safe_text(font->t248.t7);
		printf("\n");
	}

	if (fprintf(fp, "\n") < 0) {
		return -1;
	}

	return 0;
}

static int print_safe_text(const char *text)
{
	if (text != NULL && *text != 0) {
		if (printf("%s", text) < 0) {
			return -1;
		}
	}

	return 0;
}

static int write_safe_text(FILE *fp, const char *text)
{
	if (text != NULL && *text != 0) {
		if (fprintf(fp, "%s", text) < 0) {
			return -1;
		}
	}

	return 0;
}

static void free_font_info(struct font_t141_t *font)
{
	safe_free(font->t12);
	safe_free(font->t14);
	safe_free(font->t55);
	safe_free(font->t703);
	safe_free(font->t635);
	safe_free(font->t249);
	safe_free(font->t250);
	safe_free(font->t540);
	safe_free(font->t634);
	safe_free(font->t623);
	safe_free(font->t143.t22);
	safe_free(font->t248.t12);
	safe_free(font->t248.t32);
	safe_free(font->t248.t7);
	safe_free(font->t248.t50);
	safe_free(font->t248.t8.t1);
	safe_free(font->t248.t8.t2);

	memset(font, 0, sizeof(*font));
}

static void safe_free(void *ptr)
{
	if (ptr != NULL) {
		free(ptr);
	}
}

/* EOF */
