#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "utils.h"
#include "file_utils.h"
#ifdef WIN32
#include "my_printf.h"
#define	printf	my_printf
#endif


struct script_table_t {
	int size;
	struct script_info_t {
		char *filename;
		long id;
		char *name;
		char *text;
		int used;
	} *table;
};

struct text_table_t {
	int size;
	struct text_data_t {
		long id;
		char *en_text;
		char *ja_text;
		char *ja_text2;
		int used;
	} *table;
};

#if 1
static const char *ja_char_set = NULL;
#else
static const char *ja_char_set = "!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~ " \
	"§¨°±´×÷‐―‘’“”‥…′″※℃←↑→↓∞∥∴≠≦≧■□▲△▼▽◆◇○◎●★☆♀♂　、。〃々〆〇〈〉《》「」『』【】〔〕" \
	"ぁあぃいぅうぇえぉおかがきぎくぐけげこごさざしじすずせぜそぞただちぢっつづてでとどなにぬねのはばぱひびぴふぶぷへべぺほぼ" \
	"ぽまみむめもゃやゅゆょよらりるれろゎわゐゑをん゛゜ゝゞァアィイゥウェエォオカガキギクグケゲコゴサザシジスズセゼソゾタダチ" \
	"ヂッツヅテデトドナニヌネノハバパヒビピフブプヘベペホボポマミムメモャヤュユョヨラリルレロヮワヰヱヲンヴヵヶ・ーヽヾ一丁七" \
	"万丈三上下不与丑且世丘丙丞両並中串丸丹主乃久之乍乎乏乗乙九乞也乱乳乾亀了予争事二云互五井亘亙些亜亡交亥亦亨享京亭亮人什仁" \
	"仇今介仏仔仕他付仙仝代令以仮仰仲件任企伊伍伎伏伐休会伝伯伴伶伸伺似伽佃但位低住佐佑体何余作佳併佼使侃例侍供依侠価侭侮侯侵" \
	"侶便係促俄俊俗保信俣修俳俵俸俺倉個倍倒倖候借倣値倦倫倭倶倹偉偏停健偲側偵偶偽傍傑傘備催傭債傷傾僅働像僑僕僚僧僻儀億儒償優" \
	"儲允元兄充兆兇先光克免兎児党兜入全八公六共兵其具典兼内円冊再冒冗写冠冥冨冬冴冶冷凄准凋凌凍凝凡処凧凪凱凶凸凹出函刀刃分切" \
	"刈刊刑列初判別利到制刷券刺刻剃則削前剖剛剣剤剥副剰割創劃劇劉力功加劣助努劫励労効劾勃勅勇勉動勘務勝募勢勤勧勲勺勾勿匁匂包" \
	"化北匙匝匠匡匪匹区医匿十千升午半卑卒卓協南単博卜占卦卯印危即却卵卸卿厄厘厚原厨厩厭厳去参又叉及友双反収叔取受叙叛叡叢口古" \
	"句叩只叫召可台叱史右叶号司吃各合吉吊吋同名后吏吐向君吟吠否含吸吹吻吾呂呆呈呉告呑周呪味呼命咋和咲咳咽哀品哉員哨哩哲唄唆唇" \
	"唐唖唯唱唾啄商問啓善喉喋喚喜喝喧喪喫喬喰営嗣嘆嘉嘗嘘嘩嘱噂噌噛器噴噸噺嚇嚢囚四回因団困囲図固国圃圏園土圧在圭地坂均坊坐坑" \
	"坤坦坪垂型垢垣埋城埜域埠埴執培基埼堀堂堅堆堕堤堪堰報場堵堺塀塁塊塑塔塗塘塙塚塞塩填塵塾境墓増墜墨墳墾壁壇壊壌壕士壬壮声壱" \
	"売壷変夏夕外夙多夜夢大天太夫央失夷奄奇奈奉奏契奔套奥奨奪奮女奴好如妃妄妊妓妖妙妥妨妬妹妻妾姉始姐姑姓委姥姦姪姫姶姻姿威娃" \
	"娘娠娩娯娼婁婆婚婦婿媒媛嫁嫉嫌嫡嬉嬢嬬嬰子孔字存孜孝孟季孤学孫宅宇守安宋完宍宏宕宗官宙定宛宜宝実客宣室宥宮宰害宴宵家容宿" \
	"寂寄寅密富寒寓寛寝察寡寧審寮寵寸寺対寿封専射将尉尊尋導小少尖尚尤尭就尺尻尼尽尾尿局居屈届屋屍屑展属屠屡層履屯山岐岡岨岩岬" \
	"岱岳岸峠峡峨峯峰島峻崇崎崖崩嵐嵩嵯嶋嶺巌川州巡巣工左巧巨差己巳巴巷巻巽巾市布帆希帖帝帥師席帯帰帳常帽幅幌幕幡幣干平年幸幹" \
	"幻幼幽幾庁広庄庇床序底庖店庚府度座庫庭庵庶康庸廃廉廊廓廟廠延廷建廻廼廿弁弄弊式弐弓弔引弗弘弛弟弥弦弧弱張強弼弾彊当形彦彩" \
	"彪彫彬彰影役彼往征径待律後徐徒従得御復循微徳徴徹徽心必忌忍志忘忙応忠快念忽怒怖怜思怠急性怨怪怯恋恐恒恕恢恥恨恩恭息恰恵悉" \
	"悌悔悟悠患悦悩悪悲悶悼情惇惑惚惜惟惣惨惰想惹愁愈愉意愚愛感慈態慌慎慕慢慣慧慨慮慰慶慾憂憎憐憤憧憩憲憶憾懇懐懲懸戊戎成我戒" \
	"或戚戟戦戯戴戸戻房所扇扉手才打払托扮扱扶批承技抄把抑投抗折抜択披抱抵抹押抽担拍拐拒拓拘拙招拝拠拡括拭拳拶拷拾持指按挑挙挟" \
	"挨挫振挺挽挿捉捌捕捗捜捧捨据捲捷捺捻掃授掌排掘掛掠採探接控推掩措掬掲掴掻揃描提揖揚換握揮援揺損搬搭携搾摂摘摩摸摺撃撒撚撞" \
	"撤撫播撮撰撲撹擁操擢擦擬擾支改攻放政故敏救敗教敢散敦敬数整敵敷文斉斌斎斐斑斗料斜斡斤斥斧斬断斯新方於施旅旋族旗既日旦旧旨" \
	"早旬旭旺昂昆昇昌明昏易昔星映春昧昨昭是昼時晃晋晒晦晩普景晴晶智暁暇暑暖暗暢暦暫暮暴曇曙曜曝曲曳更書曹曽曾替最月有朋服朔朕" \
	"朗望朝期木未末本札朱朴机朽杉李杏材村杓杖杜束条杢来杭杯東杵杷松板枇析枕林枚果枝枠枢枯架柁柄柊柏某柑染柔柘柚柱柳柴柵査柾柿" \
	"栂栃栄栓栖栗校栢株栴核根格栽桁桂桃案桐桑桓桔桜桝桟桧桶梁梅梓梗梢梧梨梯械梱梶梼棄棉棋棒棚棟森棲棺椀椅椋植椎椙椛検椴椿楊楓" \
	"楕楚楠楢業楯楳極楼楽概榊榎榔榛構槌槍様槙槻槽樋樗標樟模権横樫樵樹樺樽橋橘機橡橿檀檎櫓櫛櫨欄欝欠次欣欧欲欺欽款歌歎歓止正此" \
	"武歩歪歯歳歴死殆殉殊残殖殴段殺殻殿毅母毎毒比毘毛氏民気水氷永氾汀汁求汎汐汗汚汝江池汰汲決汽沃沈沌沓沖沙没沢沫河沸油治沼沿" \
	"況泉泊泌法泡波泣泥注泰泳洋洗洛洞津洩洪洲活派流浄浅浜浦浩浪浬浮浴海浸消涌涙涛涜涯液涼淀淋淑淘淡淫深淳淵混添清渇済渉渋渓渚" \
	"減渠渡渥渦温測港湊湖湘湛湧湯湾湿満溌源準溜溝溢溶溺滅滋滑滝滞滴漁漂漆漉漏演漕漠漢漣漫漬漸潅潔潜潟潤潮潰澄澗澱激濁濃濠濡濫" \
	"濯瀕瀞瀦瀧瀬灘火灯灰灸灼災炉炊炎炭点為烈烏烹焔焚無焦然焼煉煎煙煤照煩煮煽熊熔熟熱燃燈燐燕燥燦燭爆爪爵父爺爽爾片版牌牒牙牛" \
	"牝牟牡牢牧物牲特牽犀犠犬犯状狂狐狗狙狛狩独狭狸狼狽猛猟猪猫献猶猷猿獄獅獣獲玄率玉王玖玩玲珂珊珍珠珪班現球理琉琢琳琴琵琶瑚" \
	"瑛瑞瑠瑳璃環璽瓜瓢瓦瓶甑甘甚甜生産甥用甫田由甲申男町画界畏畑畔留畜畝畠畢略畦番異畳畷畿疋疎疏疑疫疲疹疾病症痔痕痘痛痢痩痴" \
	"療癌癒癖発登白百的皆皇皐皮皿盃盆盈益盗盛盟監盤目盲直相盾省眉看県真眠眺眼着睡督睦瞥瞬瞭瞳矛矢知矧矩短矯石砂研砕砥砦砧砲破" \
	"砺砿硝硫硬硯硲碁碇碍碑碓碕碗碧碩確磁磐磨磯礁礎示礼社祁祇祈祉祐祖祝神祢祥票祭祷禁禄禅禍禎福禦禰禽禾禿秀私秋科秒秘租秤秦秩" \
	"称移稀程税稔稗稚稜種稲稼稽稿穀穂穆積穎穏穐穣穫穴究空穿突窃窄窒窓窟窪窮窯窺竃立竜章竣童竪端競竹竺竿笈笑笛笠笥符第笹筆筈等" \
	"筋筏筑筒答策箆箇箔箕算管箪箭箱箸節範篇築篠篤篭簡簸簾簿籍米籾粁粂粉粋粍粒粕粗粘粛粟粥粧精糊糎糖糞糟糠糧糸系糾紀約紅紋納紐" \
	"純紗紘紙級紛素紡索紫紬累細紳紹紺終絃組経結絞絡絢給統絵絶絹継続綜綬維綱網綴綻綾綿緊緋総緑緒線締編緩緬緯練縁縄縛縞縦縫縮績" \
	"繁繊繋繍織繕繭繰纂纏缶罪罫置罰署罵罷羅羊美群羨義羽翁翌習翠翫翰翻翼耀老考者而耐耕耗耳耶耽聖聞聡聯聴職聾肇肉肋肌肖肘肝股肢" \
	"肥肩肪肯肱育肴肺胃胆背胎胞胡胤胴胸能脂脅脆脇脈脊脚脱脳脹腎腐腔腕腫腰腸腹腺腿膏膚膜膝膨膳膿臆臓臣臥臨自臭至致臼興舌舎舗舘" \
	"舛舜舞舟航般舵舶舷船艇艦艮良色艶芋芙芝芥芦芭芯花芳芸芹芽苅苑苓苔苗苛若苦苧苫英茂茄茅茎茜茨茶茸草荊荏荒荘荷荻莞莫莱菅菊菌" \
	"菓菖菜菟菩華菰菱萄萌萎萩萱落葉葎著葛葡董葦葬葱葵葺蒋蒐蒔蒙蒜蒲蒸蒼蓄蓉蓋蓑蓬蓮蔀蔑蔓蔚蔦蔭蔵蔽蕃蕉蕊蕎蕗蕨蕩蕪薄薗薙薦薩" \
	"薪薫薬薮薯藁藍藤藩藷藻蘇蘭虎虐虚虜虞虫虹虻蚊蚕蚤蛇蛋蛍蛎蛙蛤蛭蛮蛸蛾蜂蜘蜜蝉蝋蝕蝦蝶蝿融螺蟹蟻血衆行術街衛衝衡衣表衰衷衿" \
	"袈袋袖被袴袷裁裂装裏裕補裟裡裳裸製裾複褐褒襖襟襲西要覆覇見規視覗覚覧親観角解触言訂計訊討訓託記訟訣訪設許訳訴診註証詐詑詔" \
	"評詞詠詣試詩詫詮詰話該詳誇誉誌認誓誕誘語誠誤説読誰課誹誼調談請諌諏諒論諜諦諭諮諸諺諾謀謁謂謄謎謙講謝謡謬謹識譜警議譲護讃" \
	"讐谷豆豊豚象豪豹貌貝貞負財貢貧貨販貫責貯貰貴買貸費貼貿賀賂賃賄資賊賎賑賓賛賜賞賠賢賦質賭購贈贋赤赦赫走赴起超越趣趨足距跡" \
	"跨路跳践踊踏蹄蹟蹴躍身躯車軌軍軒軟転軸軽較載輔輝輩輪輯輸輿轄轍轟轡辛辞辰辱農辺辻込辿迂迄迅迎近返迦迩迫迭述迷追退送逃逆透" \
	"逐逓途逗這通逝速造逢連逮週進逸逼遁遂遅遇遊運遍過道達違遜遠遡遣遥適遭遮遵遷選遺遼避還邑那邦邪邸郁郊郎郡部郭郵郷都鄭酉酋酌" \
	"配酎酒酔酢酪酬酵酷酸醇醍醐醒醗醜醤醸釆采釈里重野量金釘釜針釣釦釧鈍鈎鈴鈷鉄鉛鉢鉦鉱鉾銀銃銅銑銘銚銭鋒鋤鋪鋭鋲鋳鋸鋼錆錐錘" \
	"錠錦錨錫錬錯録鍋鍍鍔鍛鍬鍵鍾鎌鎖鎗鎚鎧鎮鏑鏡鐘鐙鐸鑑鑓長門閃閉開閏閑間関閣閤閥閲闇闘阜阪防阻阿陀附降限陛院陣除陥陪陰陳陵" \
	"陶陸険陽隅隆隈隊階随隔隙際障隠隣隷隻隼雀雁雄雅集雇雌雑雛離難雨雪雫雰雲零雷電需震霊霜霞霧露青靖静非面革靭靴鞄鞍鞘鞠鞭韓韮" \
	"音韻響頁頂頃項順須預頑頒頓頗領頚頬頭頴頻頼題額顎顔顕願顛類顧風飛食飢飯飲飴飼飽飾餅養餌餐餓館饗首香馨馬馳馴駁駄駅駆駈駐駒" \
	"駕駿騎騒験騨騰驚骨骸髄高髪髭鬼魁魂魅魔魚魯鮎鮒鮪鮫鮭鮮鯉鯖鯛鯨鯵鰍鰐鰭鰯鰹鰻鱈鱒鱗鳥鳩鳳鳴鳶鴇鴎鴛鴨鴫鴬鴻鵜鵠鵡鵬鶏鶴鷲" \
	"鷹鷺鹸鹿麓麗麟麦麹麺麻麿黄黍黒黙黛鼎鼓鼠鼻齢龍";
#endif


static void usage(void);
static int read_char_set_file(const char *filename);
static int modify_text_main(const char *ja_text_file, const char *in_xml_file, const char *out_xml_file, struct script_table_t *script_table, int lang_no);
static int modify_text(struct text_table_t *table, const char *in_file, const char *out_file, struct script_table_t *script_table, int lang_no);
static int read_ja_text(struct text_table_t *table, const char *filename);
static int split_text(char *buffer, char **id_str, char **en_text, char **ja_text, char **ja_text2);
static void free_text(char *text, char *buffer, size_t buffer_size);
static void init_text_data(struct text_table_t *table);
static void free_text_data(struct text_table_t *table);
static int add_text_data(struct text_table_t *table, long id, const char *en_text, const char *ja_text, const char *ja_text2);
static struct text_data_t *find_text_data(struct text_table_t *table, long id);
static struct text_data_t *find_text_data_by_en_text(struct text_table_t *table, const char *en_text);
static char *convert_text(char *text);
static char *unquote_text(const char *text);
static int check_ja_text(char *text);
static void print_not_used(const struct text_table_t *table);
static int add_script_file(struct script_table_t *table, const char *filename);
static void free_script_table(struct script_table_t *table);
static struct script_info_t *find_script(struct script_table_t *table, const char *name, int name_len);
static int read_script_file_main(struct script_table_t *table);
static int read_script_file(struct script_info_t *script);
static int get_script_id_and_name(struct script_info_t *script, char *buffer, char **next);
static char *convert_crlf_to_br(const char *buffer);
static void print_not_used_script(const struct script_table_t *table);


int main(int argc, char **argv)
{
	int num_cmd_params = 3;
	const char *text_file = NULL;
	const char *in_xml_file = NULL;
	const char *out_xml_file = NULL;
	const char *char_set_file = NULL;
	struct script_table_t script_table;

	int lang_no = 1;

	script_table.size = 0;
	script_table.table = NULL;

	while (*++argv) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "-c") == 0 || strcmp(*argv, "--charset") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: 文字セットファイル名がありません\n");
					return 1;
				}
				char_set_file = *argv;
			} else if (strcmp(*argv, "-l") == 0 || strcmp(*argv, "--lang") == 0) {
				char *endptr;
				if (*++argv == NULL) {
					printf("ERROR: 言語番号がありません\n");
					return 1;
				}
				lang_no = strtoul(*argv, &endptr, 10);
				if (endptr == NULL || *endptr != 0 || lang_no < 0) {
					printf("ERROR: 不正な言語番号です: %s\n", *argv);
					return 1;
				}
			} else if (strcmp(*argv, "-s") == 0 || strcmp(*argv, "--script") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: スクリプトファイル名がありません\n");
					return 1;
				}
				if (add_script_file(&script_table, *argv) < 0) {
					printf("ERROR: スクリプトファイル名を登録できません\n");
					return 1;
				}
			} else {
				printf("ERROR: 不正なオプションです: %s\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("ERROR: コマンドラインパラメータが多すぎます: %s\n", *argv);
				return 1;
			}
			if (num_cmd_params == 3) {
				text_file = *argv;
			} else if (num_cmd_params == 2) {
				in_xml_file = *argv;
			} else {
				out_xml_file = *argv;
			}
			num_cmd_params--;
		}
	}

	if (text_file == NULL || in_xml_file == NULL || out_xml_file == NULL) {
		usage();
		return 1;
	}

	if (char_set_file != NULL) {
		if (read_char_set_file(char_set_file) < 0) {
			printf("ERROR: 文字セットを読み込めません\n");
			return 1;
		}
	}

	if (read_script_file_main(&script_table) < 0) {
		printf("ERROR: スクリプトを読み込めません\n");
		return 1;
	}

	if (modify_text_main(text_file, in_xml_file, out_xml_file, &script_table, lang_no) < 0) {
		printf("ERROR: テキスト更新できません\n");
		return 1;
	}

	free_script_table(&script_table);

	return 0;
}

static void usage(void)
{
	printf("使用方法: modify_text [オプション] 翻訳テキスト 入力XMLファイル 出力XMLファイル\n");
	printf("機    能: XMLファイルのテキストを翻訳テキストに置き換えます\n");
	printf("オプション:\n");
	printf("    -c|--charset ファイル名\n");
	printf("                        使用可能な文字を指定したファイルから読み込み確認します\n");
	printf("                        ファイルに無い文字を使用していた場合は'_'に置き換えます\n");
	printf("    -l|--lang 言語番号  テキストの言語番号を指定します (デフォルト:1)\n");
	printf("    -s|--script  ファイル名\n");
	printf("                        指定したファイルから置き換えるスクリプトを読み込みます\n");
	printf("    -?|--help           使用方法を表示して終了します\n");
}

static int read_char_set_file(const char *filename)
{
	int result = -1;
	FILE *fp = NULL;
	struct stat st;
	long size;
	char *new_char_set = NULL;

	if (stat(filename, &st) != 0) {
		return 0;
	}

	if ((fp = fopen(filename, "rb")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	if (fseek(fp, 0L, SEEK_END) < 0) {
		printf("ERROR: %d: ファイルシークエラー: SEEK_END: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	size = ftell(fp);

	if (fseek(fp, 0L, SEEK_SET) < 0) {
		printf("ERROR: %d: ファイルシークエラー: SEEK_SET: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	if ((new_char_set = (char *)malloc(size + 1)) == NULL) {
		printf("ERROR: memory exhausted: size=%ld\n", size + 1);
		goto ERROR_EXIT;
	}

	if (fread(new_char_set, size, 1, fp) != 1) {
		printf("ERROR: %d: ファイル読み込みエラー: %s\n", errno, filename);
		goto ERROR_EXIT;
	}
	new_char_set[size] = 0;

	fclose(fp);
	fp = NULL;

	{
		char *src = new_char_set;
		char *des = src;
		const char bom[3] = { 0xEF, 0xBB, 0xBF };

		/* Byte Order Markをスキップ */
		if (3 < size && memcmp(src, bom, sizeof(bom)) == 0) {
			src += 3;
		}

		/* 空白文字を除去 */
		while (*src != 0) {
			if (*src == ' ' || *src == '\t' || *src == '\r' || *src == '\n') {
				src++;
				continue;
			}
			*des++ = *src++;
		}
		*des = 0;
	}

	ja_char_set = new_char_set;

	printf("INFORMATION: 文字セットを%sから読み込みました\n", filename);

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	if (result < 0) {
		if (new_char_set != NULL) {
			free(new_char_set);
		}
	}

	return result;
}

static int modify_text_main(const char *ja_text_file, const char *in_xml_file, const char *out_xml_file, struct script_table_t *script_table, int lang_no)
{
	int result = -1;
	struct text_table_t text_table;

	init_text_data(&text_table);

	if (read_ja_text(&text_table, ja_text_file) < 0) {
		printf("ERROR: 翻訳テキストを読み込めません\n");
		goto ERROR_EXIT;
	}

	if (modify_text(&text_table, in_xml_file, out_xml_file, script_table, lang_no) < 0) {
		printf("ERROR: XMLファイルを更新できません\n");
		goto ERROR_EXIT;
	}

	print_not_used(&text_table);

	print_not_used_script(script_table);

	result = 0;

ERROR_EXIT:
	free_text_data(&text_table);

	return result;
}

static int modify_text(struct text_table_t *table, const char *in_file, const char *out_file, struct script_table_t *script_table, int lang_no)
{
	int result = -1;
	FILE *in_fp = NULL;
	FILE *out_fp = NULL;
	long line = 0;
	char *ptr;

	int read_status;
	char *buffer = NULL;
	size_t buffer_size = 0;
	char *buffer2 = NULL;
	size_t buffer2_size = 0;

	char t13_pattern[32];

	snprintf(t13_pattern, sizeof(t13_pattern), " T13=\"%d\"", lang_no);

	if ((in_fp = fopen(in_file, "rb")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, in_file);
		goto ERROR_EXIT;
	}

	if (make_directory(out_file, 1) < 0) {
		printf("WARNING: 出力先ディレクトリを作成できません\n");
	}

	if ((out_fp = fopen(out_file, "wb")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, out_file);
		goto ERROR_EXIT;
	}

	while (1) {
		if ((read_status = read_next_line(&buffer, &buffer_size, in_fp)) < 0) {
			printf("ERROR: ファイル読み込みエラー: %s\n", in_file);
			goto ERROR_EXIT;
		} else if (read_status == 0) {
			break;
		}

		line++;

		if ((ptr = strstr(buffer, "<T255 ")) != NULL) {
			/* テキスト */
			long line2 = line;
			char *ptr2 = NULL;

			char *start;
			long id;
			int len;
			struct text_data_t *data;

			/* <T258 ... />の行を追加読み込み */
			if (buffer2 != NULL) {
				*buffer2 = 0;
			}
			while (1) {
				if ((read_status = read_append_next_line(&buffer2, &buffer2_size, in_fp)) < 0) {
					line = line2;
					printf("ERROR: ファイル読み込みエラー: %s\n", in_file);
					goto ERROR_EXIT;
				} else if (read_status == 0) {
					break;
				}
				line2++;

				if ((ptr2 = strstr(buffer2, "<T258 ")) != NULL) {
					break;
				}
				if (strstr(buffer2, "</T255>") != NULL) {
					break;
				}
			}

			if (ptr2 == NULL) {
				printf("ERROR: <T258 ... />が見つかりません: %s: line=%ld\n", in_file, line2);
				goto ERROR_EXIT;
			}

/*			if (strstr(ptr2 + 5, " T13=\"1\"") == NULL) {*/
			if (strstr(ptr2 + 5, t13_pattern) == NULL) {
				/* 対象外言語のテキスト */
				if (fwrite(buffer, strlen(buffer), 1, out_fp) != 1 ||
					fwrite(buffer2, strlen(buffer2), 1, out_fp) != 1) {
					printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
					goto ERROR_EXIT;
				}
				line = line2;
				continue;
			}

			if ((ptr = strstr(ptr + 5, " T13=\"")) == NULL) {
				printf("ERROR: テキストIDが見つかりません: %s(%ld)\n", in_file, line);
				goto ERROR_EXIT;
			}

			id = strtoul(ptr + 6, &ptr, 10);
			if (*ptr != '\"') {
				printf("ERROR: テキストIDの書式が違います: %s(%ld)\n", in_file, line);
				goto ERROR_EXIT;
			}

			if ((ptr = strstr(ptr + 1, " T256=\"")) == NULL) {
				printf("ERROR: テキストが見つかりません: %s(%ld)\n", in_file, line);
				goto ERROR_EXIT;
			}

			start = ptr + 7;

			if ((ptr = strstr(start, "\"")) == NULL) {
				printf("ERROR: テキストが\"で終っていません: %s(%ld)\n", in_file, line);
				goto ERROR_EXIT;
			}

			len = ptr - start;

			if ((data = find_text_data(table, id)) == NULL) {
				/* テキスト未登録 */
				if (0 < len) {
					printf("翻訳テキストがありません: id=%ld, text=%.*s\n", id, len, start);
				}

				if (fwrite(buffer, strlen(buffer), 1, out_fp) != 1 ||
					fwrite(buffer2, strlen(buffer2), 1, out_fp) != 1) {
					printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
					goto ERROR_EXIT;
				}
				line = line2;
				continue;
			}

			data->used++;

			if (strlen(data->en_text) != len || strncmp(data->en_text, start, len) != 0) {
				printf("WARNING: テキストの内容が違っています: %s(%ld) id=%ld\n", in_file, line, id);
				printf("\tXML : %.*s\n", (int)len, start);
				printf("\t翻訳: %s\n", data->en_text);
			}

			if (data->ja_text != NULL && strcmp(data->ja_text, "0") == 0) {
				/* 原文が同じテキストの翻訳を使用する */
				struct text_data_t *data2;
				if ((data2 = find_text_data_by_en_text(table, data->en_text)) != NULL) {
					data = data2;
#if 0
				} else {
					printf("WARNING: 原文が同じテキストが見つかりません: %s(%ld) id=%ld\n", in_file, line, id);
#endif
				}
			}

			if ((data->ja_text == NULL && data->ja_text2 == NULL) ||
				(data->ja_text != NULL && (strcmp(data->ja_text, "原文") == 0 ||
										   ((strcmp(data->ja_text, "0") == 0 || strcmp(data->ja_text, "機械翻訳") == 0) && data->ja_text2 == NULL)))) {
				/* 原文を使用、変更しない */
				if (fwrite(buffer, strlen(buffer), 1, out_fp) != 1 ||
					fwrite(buffer2, strlen(buffer2), 1, out_fp) != 1) {
					printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
					goto ERROR_EXIT;
				}
				line = line2;
				continue;
			}

#if 0
			printf("DEBUG: ja_text=%s\n", data->ja_text);
#endif

			/* テキストより前を書き込み */
			if (fwrite(buffer, start - buffer, 1, out_fp) != 1) {
				printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
				goto ERROR_EXIT;
			}

			/* 翻訳テキストを書き込み */
			if (data->ja_text == NULL || strcmp(data->ja_text, "0") == 0 || strcmp(data->ja_text, "機械翻訳") == 0) {
				if (fwrite(data->ja_text2, strlen(data->ja_text2), 1, out_fp) != 1) {
					printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
					goto ERROR_EXIT;
				}
			} else {
				if (fwrite(data->ja_text, strlen(data->ja_text), 1, out_fp) != 1) {
					printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
					goto ERROR_EXIT;
				}
			}

			/* テキストより後を書き込み */
			if (fwrite(ptr, strlen(ptr), 1, out_fp) != 1 ||
				fwrite(buffer2, strlen(buffer2), 1, out_fp) != 1) {
				printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
				goto ERROR_EXIT;
			}

			line = line2;

		} else if (script_table->size != 0 && (ptr = strstr(buffer, "<T650 ")) != NULL) {
			/* スクリプト */
			char *name;
			long id;
			struct script_info_t *script;

			if ((ptr = strstr(ptr + 5, " T12=\"")) == NULL) {
				printf("ERROR: スクリプト名が見つかりません: %s(%ld)\n", in_file, line);
				goto ERROR_EXIT;
			}

			name = ptr + 6;
			if ((ptr = strstr(name, "\"")) == NULL) {
				printf("ERROR: スクリプト名が\"で終っていません: %s(%ld)\n", in_file, line);
				goto ERROR_EXIT;
			}

			if ((script = find_script(script_table, name, ptr - name)) == NULL) {
				/* スクリプト名不一致 */
				if (fwrite(buffer, strlen(buffer), 1, out_fp) != 1) {
					printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
					goto ERROR_EXIT;
				}
				continue;
			}

			if ((ptr = strstr(ptr + 1, " T13=\"")) == NULL) {
				printf("ERROR: スクリプトIDが見つかりません: %s(%ld)\n", in_file, line);
				goto ERROR_EXIT;
			}

			id = strtoul(ptr + 6, &ptr, 10);
			if (ptr == NULL || *ptr != '\"') {
				printf("ERROR: スクリプトIDの書式が違います: %s(%ld)\n", in_file, line);
				goto ERROR_EXIT;
			}

			if (id != script->id) {
				/* スクリプトID不一致 */
				printf("WARNING: スクリプト名は一致しましたが、IDは不一致です ... 置換しません: %s(%ld): name=%s, id=%ld\n", in_file, line, script->name, id);

				if (fwrite(buffer, strlen(buffer), 1, out_fp) != 1) {
					printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
					goto ERROR_EXIT;
				}
				continue;
			}

			if ((ptr = strstr(ptr + 1, " T652=\"")) == NULL) {
				printf("ERROR: スクリプト本体が見つかりません: %s(%ld)\n", in_file, line);
				goto ERROR_EXIT;
			}

			ptr += 7;

			/* スクリプトより前を書き込み */
			if (fwrite(buffer, ptr - buffer, 1, out_fp) != 1) {
				printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
				goto ERROR_EXIT;
			}

			/* スクリプトを書き込み */
			if (fwrite(script->text, strlen(script->text), 1, out_fp) != 1) {
				printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
				goto ERROR_EXIT;
			}

			if ((ptr = strstr(ptr, "\"")) == NULL) {
				printf("ERROR: スクリプトが\"で終っていません: %s(%ld)\n", in_file, line);
				goto ERROR_EXIT;
			}

			/* スクリプトより後を書き込み */
			if (fwrite(ptr, strlen(ptr), 1, out_fp) != 1) {
				printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
				goto ERROR_EXIT;
			}

			script->used++;
		} else {
			/* 書き換え対象以外 */
			if (fwrite(buffer, strlen(buffer), 1, out_fp) != 1) {
				printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
				goto ERROR_EXIT;
			}
		}
	}

	result = 0;

ERROR_EXIT:
	if (in_fp != NULL) {
		fclose(in_fp);
	}
	if (out_fp != NULL) {
		fclose(out_fp);
	}
	if (buffer != NULL) {
		free(buffer);
	}
	if (buffer2 != NULL) {
		free(buffer2);
	}

	return result;
}

static int read_ja_text(struct text_table_t *table, const char *filename)
{
	int result = -1;
	FILE *fp = NULL;
	long line = 0;

	int read_status;
	char *buffer = NULL;
	size_t buffer_size = 0;

	if ((fp = fopen(filename, "rt")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	while (1) {
		long id;
		char *id_str;
		char *en_text;
		char *ja_text;
		char *ja_text2;
		char *endptr;

		if ((read_status = read_next_line(&buffer, &buffer_size, fp)) < 0) {
			printf("ERROR: ファイル読み込みエラー: %s\n", filename);
			goto ERROR_EXIT;
		} else if (read_status == 0) {
			break;
		}

		line++;
		if (line == 1) {
			continue;
		}

		if (split_text(buffer, &id_str, &en_text, &ja_text, &ja_text2) < 0) {
			printf("ERROR: ファイル形式が違います: %s(%ld)\n", filename, line);
			goto ERROR_EXIT;
		}

		id = strtoul(id_str, &endptr, 10);
		if (endptr == NULL || *endptr != 0) {
			printf("ERROR: ファイル形式が違います: %s(%ld)\n", filename, line);
			goto ERROR_EXIT;
		}

		if ((en_text = convert_text(en_text)) == NULL) {
			printf("ERROR: 文字列変換できません\n");
			goto ERROR_EXIT;
		}
		if ((ja_text = convert_text(ja_text)) == NULL) {
			printf("ERROR: 文字列変換できません\n");
			goto ERROR_EXIT;
		}
		if ((ja_text2 = convert_text(ja_text2)) == NULL) {
			printf("ERROR: 文字列変換できません\n");
			goto ERROR_EXIT;
		}

		if (((strcmp(ja_text, "0") != 0 && strcmp(ja_text, "機械翻訳") != 0 && strcmp(ja_text, "原文") != 0) && check_ja_text(ja_text) < 0) ||
			check_ja_text(ja_text2) < 0) {
			printf("WARNING: 表示できない文字が使用されています: %s(%ld)\n", filename, line);
		}

		if (add_text_data(table, id, en_text, ja_text, ja_text2) < 0) {
			printf("ERROR: テキスト登録できません\n");
			free_text(en_text, buffer, buffer_size);
			free_text(ja_text, buffer, buffer_size);
			free_text(ja_text2, buffer, buffer_size);
			goto ERROR_EXIT;
		}

		free_text(en_text, buffer, buffer_size);
		free_text(ja_text, buffer, buffer_size);
		free_text(ja_text2, buffer, buffer_size);
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (buffer != NULL) {
		free(buffer);
	}

	return result;
}

static int split_text(char *buffer, char **id_str, char **en_text, char **ja_text, char **ja_text2)
{
	char *ptr = buffer;

	*id_str = *en_text = *ja_text = *ja_text2 = NULL;

	*id_str = buffer;
	if ((ptr = strchr(ptr, '\t')) == NULL) {
		return -1;
	}
	*ptr++ = 0;

	*en_text = ptr;
	if ((ptr = strchr(ptr, '\t')) == NULL) {
		return -1;
	}
	*ptr++ = 0;

	*ja_text = ptr;
	if ((ptr = strchr(ptr, '\t')) == NULL) {
		if ((ptr = strchr(*ja_text, '\r')) != NULL) {
			*ptr = 0;
		}
		if ((ptr = strchr(*ja_text, '\n')) != NULL) {
			*ptr = 0;
		}

		*ja_text2 = *ja_text + strlen(*ja_text);	/* 空文字列 */
	} else {
		*ptr++ = 0;
		*ja_text2 = ptr;

		if ((ptr = strchr(*ja_text2, '\t')) != NULL) {
			*ptr = 0;
		}
		if ((ptr = strchr(*ja_text2, '\r')) != NULL) {
			*ptr = 0;
		}
		if ((ptr = strchr(*ja_text2, '\n')) != NULL) {
			*ptr = 0;
		}
	}

	return 0;
}

static void free_text(char *text, char *buffer, size_t buffer_size)
{
	if (text == NULL || (buffer <= text && text < buffer + buffer_size)) {
		return;
	}
	free(text);
}

static void init_text_data(struct text_table_t *table)
{
	table->size = 0;
	table->table = NULL;
}

static void free_text_data(struct text_table_t *table)
{
	int i;

	for (i = 0; i < table->size; i++) {
		if (table->table[i].en_text != NULL) {
			free(table->table[i].en_text);
		}
		if (table->table[i].ja_text != NULL) {
			free(table->table[i].ja_text);
		}
	}

	free(table->table);
	table->table = NULL;
	table->size = 0;
}

static int add_text_data(struct text_table_t *table, long id, const char *en_text, const char *ja_text, const char *ja_text2)
{
	int i;

	for (i = 0; i < table->size; i++) {
		if (table->table[i].id == id) {
			printf("WARNING: テキストは登録済みです ... 上書きします: id=%ld\n", id);
			break;
		}
		if (table->table[i].id < 0) {
			break;
		}
	}

	if (table->size <= i) {
		int new_size = table->size + 500;
		long alloc_size = sizeof(struct text_data_t) * new_size;

		if (table->table == NULL) {
			table->table = (struct text_data_t *)malloc(alloc_size);
		} else {
			table->table = (struct text_data_t *)realloc(table->table, alloc_size);
		}
		if (table->table == NULL) {
			printf("ERROR: memory exhausted: size=%ld\n", alloc_size);
			return -1;
		}

		for (i = table->size; i < new_size; i++) {
			table->table[i].id = -1;
			table->table[i].en_text = NULL;
			table->table[i].ja_text = NULL;
			table->table[i].ja_text2 = NULL;
			table->table[i].used = 0;
		}

		i = table->size;
		table->size = new_size;
	}

	table->table[i].id = id;

	if (en_text != NULL) {
		if (table->table[i].en_text != NULL) {
			free(table->table[i].en_text);
			table->table[i].en_text = NULL;
		}
		if ((table->table[i].en_text = strdup(en_text)) == NULL) {
			printf("ERROR: memory exhausted: strdup(%s)\n", en_text);
			return -1;
		}
	}

	if (ja_text != NULL && *ja_text != 0) {
		if (table->table[i].ja_text != NULL) {
			free(table->table[i].ja_text);
			table->table[i].ja_text = NULL;
		}
		if ((table->table[i].ja_text = strdup(ja_text)) == NULL) {
			printf("ERROR: memory exhausted: strdup(%s)\n", ja_text);
			return -1;
		}
	}

	if (ja_text2 != NULL && *ja_text2 != 0) {
		if (table->table[i].ja_text2 != NULL) {
			free(table->table[i].ja_text2);
			table->table[i].ja_text2 = NULL;
		}
		if ((table->table[i].ja_text2 = strdup(ja_text2)) == NULL) {
			printf("ERROR: memory exhausted: strdup(%s)\n", ja_text2);
			return -1;
		}
	}

	return 1;
}

static struct text_data_t *find_text_data(struct text_table_t *table, long id)
{
	int i;

	for (i = 0; i < table->size; i++) {
		if (table->table[i].id == id) {
			return &(table->table[i]);
		}
	}

	return NULL;
}

static struct text_data_t *find_text_data_by_en_text(struct text_table_t *table, const char *en_text)
{
	int i;
	struct text_data_t *save = NULL;

	for (i = 0; i < table->size; i++) {
		if (table->table[i].id < 0 ||
			strcmp(table->table[i].en_text, en_text) != 0) {
			continue;
		}
		if (table->table[i].ja_text == NULL &&
			table->table[i].ja_text2 != NULL &&
			save == NULL) {
			save = &(table->table[i]);
		}
		if (table->table[i].ja_text == NULL ||
			table->table[i].ja_text[0] == 0 ||
			strcmp(table->table[i].ja_text, "0") == 0 ||
			strcmp(table->table[i].ja_text, "機械翻訳") == 0) {
			continue;
		}

		return &(table->table[i]);
	}

	return save;
}

static char *convert_text(char *text)
{
	const char *orig_text = text;
	char *new_text = NULL;

	if (*text == '\"' && *(text + 1) != 0 && *(text + strlen(text) - 1) == '\"') {
		if ((new_text = unquote_text(text)) == NULL) {
			printf("ERROR: \"を除去できません\n");
			goto ERROR_EXIT;
		}
		text = new_text;
	}

	if (check_escape_chars(text)) {
		if ((new_text = escape_text(text)) == NULL) {
			printf("ERROR: 実体参照に変換できません\n");
			goto ERROR_EXIT;
		}

		if (text != orig_text) {
			free(text);
		}
		text = new_text;
	}

	return text;

ERROR_EXIT:
	if (text != orig_text) {
		free(text);
	}
	return NULL;
}

static char *unquote_text(const char *text)
{
	char *new_text = NULL;
	const char *src;
	char *des;
	int alloc_size;

	if (*text != '\"') {
		printf("ERROR: 文字列が\"で初まっていません: %s\n", text);
		goto ERROR_EXIT;
	}

	alloc_size = strlen(text) + 1;
	if ((new_text = (char *)malloc(alloc_size)) == NULL) {
		printf("ERROR: memory exhausted: size=%d\n", alloc_size);
		goto ERROR_EXIT;
	}

	src = text;
	des = new_text;

	if (*src == '\"') {
		src++;
	}

	while (*src != 0) {
		if (*src == '\"') {
			if (*(src + 1) == '\"') {
				*des++ = *src++;
				src++;
			} else {
				break;
			}
		} else {
			*des++ = *src++;
		}
	}

	*des++ = 0;

	if (*src != '\"') {
		printf("ERROR: \"で初まる文字列の最後に\"がありません: %s\n", text);
		goto ERROR_EXIT;
	}
	src++;

	if (*src != 0) {
		printf("ERROR: \"で初まる文字列の途中にエスケープされていない\"があります: %s\n", text);
		goto ERROR_EXIT;
	}

	return new_text;

ERROR_EXIT:
	if (new_text != NULL) {
		free(new_text);
	}

	return NULL;
}

static int check_ja_text(char *text)
{
	int result = 0;
	unsigned long code1;
	unsigned long code2;
	char *des = text;
	int len = 0;

	if (text == NULL || *text == 0) {
		return 0;
	}

	for ( ; *text != 0; text += len) {
		if ((len = get_next_utf8_char(&code1, text, 1)) < 0) {
			printf("WARNING: 不正な文字です: %s\n", text);
			result = -1;
			goto ERROR_EXIT;
		}
		if (code1 == 0) {
			break;
		}
		if (code1 == ' ') {
			/* 半角空白 */
			*des++ = ' ';
			continue;
		}

		if (ja_char_set != NULL) {
			/* 文字チェック */
			const char *set_ptr = ja_char_set;
			int found = 0;
			while (1) {
				int len2;
				if ((len2 = get_next_utf8_char(&code2, set_ptr, 0)) < 0) {
					printf("BUG: 不正な文字です: %s\n", set_ptr);
					result = -1;
					goto ERROR_EXIT;
				}
				if (code2 == 0) {
					break;
				}
				if (code1 == code2) {
					found = 1;
					break;
				}
				set_ptr += len2;
			}

			if (! found) {
				printf("WARNING: フォントに無い文字です: %.*s\n", len, text);
				result = -1;
				*des++ = '_';		/* 半角アンダーバーに置き替え */
			} else if (des < text) {
				while (0 < len) {
					*des++ = *text++;
					len--;
				}
			} else {
				des += len;
			}
		} else {
			/* 文字チェックしない */
			if (des < text) {
				while (0 < len) {
					*des++ = *text++;
					len--;
				}
			} else {
				des += len;
			}
		}
	}

ERROR_EXIT:
	*des = 0;

	return result;
}

static void print_not_used(const struct text_table_t *table)
{
	int i;

	for (i = 0; i < table->size; i++) {
		if (table->table[i].id < 0 || table->table[i].used != 0) {
			continue;
		}

		printf("NOT USED: id=%ld\n", table->table[i].id);
		printf("\ten=%s\n", table->table[i].en_text);
		printf("\tja=%s\n", (table->table[i].ja_text == NULL ? "" : table->table[i].ja_text));
	}
}

static int add_script_file(struct script_table_t *table, const char *filename)
{
	int i;
	long alloc_size;

	for (i = 0; i < table->size; i++) {
		if (strcmp(table->table[i].filename, filename) == 0) {
			printf("WARNING: スクリプトファイル名は既に登録済みです ... 登録しません: %s\n", filename);
			return 0;
		}
	}

	alloc_size = sizeof(struct script_info_t) * (table->size + 1);
	if (table->size == 0) {
		table->table = (struct script_info_t *)malloc(alloc_size);
	} else {
		table->table = (struct script_info_t *)realloc(table->table, alloc_size);
	}
	if (table->table == NULL) {
		printf("ERROR: memory exhausted: size=%ld\n", alloc_size);
		return -1;
	}

	if ((table->table[table->size].filename = strdup(filename)) == NULL) {
		printf("ERROR: memory exhasuted: strdup(%s)\n", filename);
		return -1;
	}
	table->table[table->size].id = -1;
	table->table[table->size].name = NULL;
	table->table[table->size].text = NULL;
	table->table[table->size].used = 0;

	table->size++;

	return 1;
}

static void free_script_table(struct script_table_t *table)
{
	if (table->table != NULL) {
		int i;

		for (i = 0; i < table->size; i++) {
			if (table->table[i].filename != NULL) {
				free(table->table[i].filename);
				table->table[i].filename = NULL;
			}

			table->table[i].id = -1;

			if (table->table[i].name != NULL) {
				free(table->table[i].name);
				table->table[i].name = NULL;
			}

			if (table->table[i].text != NULL) {
				free(table->table[i].text);
				table->table[i].text = NULL;
			}

			table->table[i].used = 0;
		}

		free(table->table);
	}

	table->size = 0;
	table->table = NULL;
}

static struct script_info_t *find_script(struct script_table_t *table, const char *name, int name_len)
{
	int i;

	for (i = 0; i < table->size; i++) {
		if (strlen(table->table[i].name) != name_len) {
			continue;
		}
		if (strncmp(table->table[i].name, name, name_len) != 0) {
			continue;
		}

		return table->table + i;
	}

	return NULL;
}

static int read_script_file_main(struct script_table_t *table)
{
	int i;

	for (i = 0; i < table->size; i++) {
		if (read_script_file(table->table + i) < 0) {
			printf("ERROR: スクリプトを読み込めません: %s\n", table->table[i].filename);
			return -1;
		}
	}

	return 0;
}

static int read_script_file(struct script_info_t *script)
{
	int result = -1;
	FILE *fp = NULL;
	long size;
	char *buffer = NULL;
	char *buffer2 = NULL;
	char *ptr;

	if ((fp = fopen(script->filename, "rb")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, script->filename);
		goto ERROR_EXIT;
	}

	if (fseek(fp, 0L, SEEK_END) < 0) {
		printf("ERROR: %d: ファイルシークエラー: SEEK_END: %s\n", errno, script->filename);
		goto ERROR_EXIT;
	}

	size = ftell(fp);

	if (fseek(fp, 0L, SEEK_SET) < 0) {
		printf("ERROR: %d: ファイルシークエラー: SEEK_SET: %s\n", errno, script->filename);
		goto ERROR_EXIT;
	}

	if ((buffer = (char *)malloc(size + 1)) == NULL) {
		printf("ERROR: memory exhausted: size=%ld\n", size + 1);
		goto ERROR_EXIT;
	}

	if (fread(buffer, size, 1, fp) != 1) {
		printf("ERROR: %d: ファイル読み込みエラー: %s\n", errno, script->filename);
		goto ERROR_EXIT;
	}
	buffer[size] = 0;

	fclose(fp);
	fp = NULL;

	if (get_script_id_and_name(script, buffer, &ptr) < 0) {
		printf("ERROR: スクリプトIDまたはスクリプト名を取得できません: %s\n", script->filename);
		goto ERROR_EXIT;
	}

	if ((buffer2 = convert_crlf_to_br(ptr)) == NULL) {
		printf("ERROR: 改行コードを<br/>に変換できません: %s\n", script->filename);
		goto ERROR_EXIT;
	}

	free(buffer);
	buffer = buffer2;
	buffer2 = NULL;

	if (check_escape_chars(buffer)) {
		if ((buffer2 = escape_text(buffer)) == NULL) {
			printf("ERROR: 実体参照に変換できません\n");
			goto ERROR_EXIT;
		}

		free(buffer);
		buffer = buffer2;
		buffer2 = NULL;
	}

	script->text = buffer;
	buffer = NULL;

	result = 0;

ERROR_EXIT:

	if (fp != NULL) {
		fclose(fp);
	}
	if (buffer != NULL) {
		free(buffer);
	}
	if (buffer2 != NULL) {
		free(buffer2);
	}

	return result;
}

static int get_script_id_and_name(struct script_info_t *script, char *buffer, char **next)
{
	const char *start;
	const char *end;
	int len;
	const char bom[3] = { 0xEF, 0xBB, 0xBF };

	if (3 <= strlen(buffer) && memcmp(buffer, bom, 3) == 0) {
		buffer += 3;
	}

	while (*buffer == ' ') {
		buffer++;
	}

	if (strncasecmp(buffer, "ID", 2) != 0) {
		printf("ERROR: スクリプトIDがありません\n");
		return -1;
	}
	buffer += 2;

	while (*buffer == ' ') {
		buffer++;
	}

	if (*buffer != '=') {
		printf("ERROR: スクリプトIDがありません\n");
		return -1;
	}
	buffer++;

	while (*buffer == ' ') {
		buffer++;
	}

	script->id = strtoul(buffer, &buffer, 10);
	if (buffer == NULL || (*buffer != ' ' && *buffer != '\r' && *buffer != '\n')) {
		printf("ERROR: 不正なスクリプトIDです\n");
		return -1;
	}

	while (*buffer == ' ') {
		buffer++;
	}

	if (*buffer == '\r') {
		buffer++;
	}
	if (*buffer != '\n') {
		printf("ERROR: 改行コードがありません\n");
		return -1;
	}
	buffer++;

	while (*buffer == ' ') {
		buffer++;
	}

	if (strncasecmp(buffer, "name", 4) != 0) {
		printf("ERROR: スクリプト名がありません\n");
		return -1;
	}
	buffer += 4;
	
	while (*buffer == ' ') {
		buffer++;
	}

	if (*buffer != '=') {
		printf("ERROR: スクリプト名がありません\n");
		return -1;
	}
	buffer++;

	while (*buffer == ' ') {
		buffer++;
	}
	start = buffer;

	while (*buffer != '\r' && *buffer != '\n') {
		buffer++;
	}
	end = buffer;

	while (start < end && *(end - 1) == ' ') {
		end--;
	}

	len = end - start;
	if (len == 0) {
		printf("ERROR: スクリプト名がありません\n");
		return -1;
	}

	if ((script->name = (char *)malloc(len + 1)) == NULL) {
		printf("ERROR: memory exhausted: size=%d\n", len + 1);
		return -1;
	}

	memcpy(script->name, start, len);
	script->name[len] = 0;

	if (*buffer == '\r') {
		buffer++;
	}
	if (*buffer != '\n') {
		printf("ERROR: 改行コードがありません\n");
		return -1;
	}
	buffer++;

	*next = buffer;

	if (check_escape_chars(script->name)) {
		char *name2;

		if ((name2 = escape_text(script->name)) == NULL) {
			printf("ERROR: 実体参照に変換できません\n");
			return -1;
		}

		free(script->name);
		script->name = name2;
	}

	return 0;
}

static char *convert_crlf_to_br(const char *buffer)
{
	long len = 0;
	const char *src = buffer;
	char *des;
	char *buffer2;

	while (*src != 0) {
		if (*src == '\n') {
			len += 5;
		} else if (*src != '\r') {
			len++;
		}
		src++;
	}

	if ((buffer2 = (char *)malloc(len + 1)) == NULL) {
		printf("ERROR: memory exhausted: size=%ld\n", len + 1);
		return NULL;
	}

	src = buffer;
	des = buffer2;

	while (*src != 0) {
		if (*src == '\n') {
			*des++ = '<';
			*des++ = 'b';
			*des++ = 'r';
			*des++ = '/';
			*des++ = '>';
		} else if (*src != '\r') {
			*des++ = *src;
		}
		src++;
	}
	*des = 0;

	return buffer2;
}

static void print_not_used_script(const struct script_table_t *table)
{
	int i;

	for (i = 0; i < table->size; i++) {
		if (table->table[i].used == 0) {
			printf("WARNING: NOT USED: script id=%ld, name=%s, file=%s\n", table->table[i].id, table->table[i].name, table->table[i].filename);
		}
	}
}

/* EOF */
