#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <iconv.h>
#include <errno.h>

#include "my_printf.h"


static int first = 1;
static int convert_to_cp932 = 0;


int my_printf(const char *format, ...)
{
	int result = -1;
	va_list ap;
	char *text = NULL;
	int text_size = 0;
	iconv_t cd = (iconv_t)-1;
	char *src;
	size_t src_left;

	if (first) {
		const char *lang;
		if ((lang = getenv("LANG")) == NULL) {
			convert_to_cp932 = 1;
		} else {
			char *dot = strrchr(lang, '.');
			if (dot != NULL) {
				lang = dot + 1;
			}
			if (strcasecmp(lang, "UTF-8") != 0) {
				convert_to_cp932 = 1;
			}
		}
		first = 0;
	}

	if (convert_to_cp932 == 0) {
		va_start(ap, format);
		vprintf(format, ap);
		va_end(ap);
		return 0;
	}

	/* 必要なバッファサイズを求める */
	va_start(ap, format);
	if ((text_size = vsnprintf(NULL, 0, format, ap)) <= 0) {
		va_end(ap);
		goto ERROR_EXIT;
	}
	va_end(ap);

	/* バッファ確保q */
	if ((text = (char *)malloc(text_size + 1)) == NULL) {
		goto ERROR_EXIT;
	}

	/* 文字列作成 */
	va_start(ap, format);
	if (vsnprintf(text, text_size + 1, format, ap) != text_size) {
		va_end(ap);
		goto ERROR_EXIT;
	}
	va_end(ap);

	/* 文字コード変換 */
	if ((cd = iconv_open("CP932", "UTF-8")) == (iconv_t)-1) {
		goto ERROR_EXIT;
	}

	src = text;
	src_left = strlen(src);
	while (0 < src_left) {
		char buffer[4096];
		char *des = buffer;
		size_t des_left = sizeof(buffer) - 1;

		if (iconv(cd, &src, &src_left, &des, &des_left) == (size_t)-1) {
			if (errno == E2BIG) {
				*des = '\0';
				fputs(buffer, stdout);
			} else if (errno == 42) {
				*des = '\0';
				fputs(buffer, stdout);

				/* 変換できない文字を ? に置換 */
				fputc('?', stdout);

				/* 置換した文字をスキップ */
				if ((*src & 0xC0) == 0xC0) {
					src++;
					src_left--;
					while (0 < src_left && (*src & 0xC0) == 0x80) {
						src++;
						src_left--;
					}
				} else {
					src++;
					src_left--;
				}
			} else {
				printf("ERROR: %d: iconv\n", errno);
				goto ERROR_EXIT;
			}
		} else {
			*des = '\0';
			fputs(buffer, stdout);
		}
	}

	result = 0;

ERROR_EXIT:
	if (cd != (iconv_t)-1) {
		iconv_close(cd);
	}

	if (text != NULL) {
		free(text);
	}

	return result;
}

/* EOF */
